import numpy as np

from pypop7.optimizers.es.es import ES


class SSAES(ES):
    """Schwefel's Self-Adaptation Evolution Strategy (SSAES).

    .. note:: `SSAES` adapts the *individual* step-sizes on-the-fly, proposed by Schwefel. Since it needs the
       *large* populations (e.g. larger than number of dimensions) for reliable adaptation, `SSAES` often
       suffers from *slow* convergence for large-scale black-box optimization (LSBBO).

       It is **highly recommended** to first attempt other more advanced ES variants for LSBBO. Here we include
       it mainly for *benchmarking* and *theoretical* purpose.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with six particular settings (`keys`):
                * 'mean'            - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'           - initial global step-size (σ), mutation strength (`float`),
                * 'n_individuals'   - number of offspring (λ: lambda), offspring population size (`int`, default:
                  `5*problem['ndim_problem']`),
                * 'n_parents'       - number of parents (μ: mu), parental population size (`int`, default:
                  `int(options['n_individuals'] / 4)`),
                * 'eta_sigma'       - learning rate of global step-size (`float`, default:
                  `1.0 / np.sqrt(problem['ndim_problem']`),
                * 'eta_axis_sigmas' - learning rate of individual step-sizes (`float`, default:
                  `1.0 / np.power(problem['ndim_problem'], 1.0 / 4.0)`).

    Examples
    --------
    Use the ES optimizer `SSAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.ssaes import SSAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> ssaes = SSAES(problem, options)  # initialize the optimizer class
       >>> results = ssaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SSAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
         * Generation 10: best_so_far_y 2.14285e-01, min(y) 2.14285e-01 & Evaluations 100
         * Generation 20: best_so_far_y 2.10049e-01, min(y) 2.10462e-01 & Evaluations 200
         ...
         * Generation 480: best_so_far_y 1.33198e-01, min(y) 1.33198e-01 & Evaluations 4800
         * Generation 490: best_so_far_y 1.31319e-01, min(y) 1.31645e-01 & Evaluations 4900
       SSAES: 5000, 0.13131869620062903

    Attributes
    ----------
    n_individuals   : `int`
                      number of offspring (λ: lambda), offspring population size.
    n_parents       : `int`
                      number of parents (μ: mu), parental population size.
    mean            : `array_like`
                      initial (starting) point, mean of Gaussian search distribution.
    sigma           : `float`
                      initial global step-size (σ), mutation strength (`float`).
    eta_sigma       : `float`
                      learning rate of global step-size.
    eta_axis_sigmas : `float`
                      learning rate of individual step-sizes.

    References
    ----------
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44
    """
    def __init__(self, problem, options):
        if options.get('n_individuals') is None:
            options['n_individuals'] = 5*problem.get('ndim_problem')
        if options.get('n_parents') is None:
            options['n_parents'] = int(options['n_individuals'] / 4)
        ES.__init__(self, problem, options)
        self.axis_sigmas = self.sigma*np.ones((self.ndim_problem,))  # individual step-sizes
        if self.eta_sigma is None:
            self.eta_sigma = 1.0 / np.sqrt(self.ndim_problem)  # learning rate of global step-size
        assert self.eta_sigma > 0, f'`self.eta_sigma` = {self.eta_sigma}, but should > 0.'
        # learning rate of individual step-sizes
        self.eta_axis_sigmas = options.get('eta_axis_sigmas', 1.0 / np.power(self.ndim_problem, 1.0 / 4.0))

    def initialize(self):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean()  # mean of Gaussian search distribution
        sigmas = np.empty((self.n_individuals, self.ndim_problem))  # individual step-sizes for all offspring
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return x, mean, sigmas, y

    def iterate(self, x=None, mean=None, sigmas=None, y=None, args=None):
        for k in range(self.n_individuals):  # sample population
            if self._check_terminations():
                return x, sigmas, y
            sigma = self.eta_sigma*self.rng_optimization.standard_normal()
            axis_sigmas = self.eta_axis_sigmas*self.rng_optimization.standard_normal((self.ndim_problem,))
            sigmas[k] = self.axis_sigmas*np.exp(axis_sigmas)*np.exp(sigma)
            x[k] = mean + sigmas[k]*self.rng_optimization.standard_normal((self.ndim_problem,))
            y[k] = self._evaluate_fitness(x[k], args)
        return x, sigmas, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, sigmas, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, sigmas, y = self.iterate(x, mean, sigmas, y, args)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            order = np.argsort(y)[:self.n_parents]
            self.axis_sigmas = np.mean(sigmas[order], axis=0)
            mean = np.mean(x[order], axis=0)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness, mean)
        results['axis_sigmas'] = self.axis_sigmas
        return results



"""
PBIL (Population-Based Incremental Learning)
Salustowicz and Schmidhuber (1997, 1998)
H6hfeld and Rudolph (1997)
Berny (2000a)
Juels (1997)

Population-based incremental learning: A method for integrating genetic search based function optimization and competitive learning
Genetic algorithms and explicit search statistics
Removing the genetics from standard genetic algorithm
Fast probabilistic modeling for combinatorial optimization
Extending population-based incremental learning to continuous search spaces
"""



"""Repeat the following paper for `CODE`:
    Wang, Y., Cai, Z., and Zhang, Q. 2011.
    Differential evolution with composite trial vector generation strategies and control parameters.
    IEEE Transactions on Evolutionary Computation, 15(1), pp.55–66.
    https://doi.org/10.1109/TEVC.2010.2087271

    Very close performance can be obtained by our code. Therefore, we argue that the repeatability of
    `CODE` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, rastrigin
from pypop7.optimizers.de.code import CODE


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 10000 * ndim_problem,
               'n_individuals': 30,
               'seed_rng': 0,  # undefined in the original paper
               'fitness_threshold': 1e-8,  # from the original paper
               }
    code = CODE(problem, options)
    results = code.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.00 vs 0.00 (from the original paper)

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5*np.ones((ndim_problem,)),
               'upper_boundary': 5*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 10000 * ndim_problem,
               'n_individuals': 30,
               'seed_rng': 0,  # undefined in the original paper
               'fitness_threshold': 1e-8,  # from the original paper
               }
    code = CODE(problem, options)
    results = code.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.00 vs 0.00 (from the original paper)



from pypop7.optimizers.eda.eda import EDA
from pypop7.optimizers.eda.umda import UMDA
from pypop7.optimizers.eda.emna import EMNA
from pypop7.optimizers.eda.aemna import AEMNA
from pypop7.optimizers.eda.rpeda import RPEDA


__all__ = [EDA,  # base (abstract) class
           EMNA, AEMNA,  # competitors for medium- and small-scale black-box optimization (often < 100)
           UMDA, RPEDA]  # EDAs especially for large-scale black-box optimization (generally >> 100)



import time
from enum import IntEnum

import numpy as np


# helper class used by all optimizer classes
class Terminations(IntEnum):
    NO_TERMINATION = 0
    MAX_FUNCTION_EVALUATIONS = 1  # maximum of function evaluations
    MAX_RUNTIME = 2  # maximal runtime
    FITNESS_THRESHOLD = 3  # stopping threshold of fitness


class Optimizer(object):
    """Base (abstract) class of all optimizers for continuous black-box minimization.

    References
    ----------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization. MIT Press.
    https://algorithmsbook.com/optimization/
    (See Chapter 7: Direct Methods for details.)

    Nesterov, Y., 2018.
    Lectures on convex optimization.
    Berlin: Springer International Publishing.
    https://link.springer.com/book/10.1007/978-3-319-91578-4

    Audet, C. and Hare, W., 2017. Derivative-free and blackbox optimization.
    Berlin: Springer International Publishing.
    https://link.springer.com/book/10.1007/978-3-319-68913-5
    """
    def __init__(self, problem, options):
        # problem-related settings
        self.fitness_function = problem.get('fitness_function')
        self.ndim_problem = problem['ndim_problem']
        self.upper_boundary = problem.get('upper_boundary')
        self.lower_boundary = problem.get('lower_boundary')
        self.initial_upper_boundary = problem.get('initial_upper_boundary', self.upper_boundary)
        self.initial_lower_boundary = problem.get('initial_lower_boundary', self.lower_boundary)
        self.problem_name = problem.get('problem_name')
        if (self.problem_name is None) and hasattr(self.fitness_function, '__name__'):
            self.problem_name = self.fitness_function.__name__
        # for function maximization, which should NOT be used by end-users (but only by developers)
        self._is_maximization = problem.get('_is_maximization', False)

        # optimizer-related options
        self.max_function_evaluations = options.get('max_function_evaluations', np.Inf)
        self.max_runtime = options.get('max_runtime', np.Inf)
        self.fitness_threshold = options.get('fitness_threshold', -np.Inf)
        if self._is_maximization:
            self.fitness_threshold *= -1.0
        self.n_individuals = options.get('n_individuals')  # offspring population size
        self.n_parents = options.get('n_parents')  # parent population size
        self.seed_rng = options.get('seed_rng')
        if self.seed_rng is None:  # it is highly recommended to explicitly set *seed_rng*
            self.rng = np.random.default_rng()  # NOT use it as much as possible
        else:
            self.rng = np.random.default_rng(self.seed_rng)
        self.seed_initialization = options.get('seed_initialization', self.rng.integers(np.iinfo(np.int64).max))
        self.rng_initialization = np.random.default_rng(self.seed_initialization)
        self.seed_optimization = options.get('seed_optimization', self.rng.integers(np.iinfo(np.int64).max))
        self.rng_optimization = np.random.default_rng(self.seed_optimization)
        self.record_fitness = options.get('record_fitness', False)
        if self.record_fitness:
            self.record_fitness_frequency = options.get('record_fitness_frequency', 1000)
        self.verbose = options.get('verbose', True)
        if self.verbose:
            self.verbose_frequency = options.get('verbose_frequency', 10)

        # auxiliary members
        self.Terminations = Terminations
        self.n_function_evaluations = options.get('n_function_evaluations', 0)
        self.start_function_evaluations = None
        self.time_function_evaluations = options.get('time_function_evaluations', 0)
        self.runtime = options.get('runtime', 0)
        self.start_time = None
        self.best_so_far_y = options.get('best_so_far_y', np.Inf)
        if self._is_maximization:
            self.best_so_far_y *= -1
        self.best_so_far_x = None
        self.termination_signal = None
        self.fitness = None
        self.is_restart = options.get('is_restart', True)

    def _evaluate_fitness(self, x, args=None):
        self.start_function_evaluations = time.time()
        if args is None:
            y = self.fitness_function(x)
        else:
            y = self.fitness_function(x, args=args)
        self.time_function_evaluations += time.time() - self.start_function_evaluations
        self.n_function_evaluations += 1
        # update best-so-far solution and fitness
        if (not self._is_maximization) and (y < self.best_so_far_y):
            self.best_so_far_x, self.best_so_far_y = np.copy(x), y
        if self._is_maximization and (-y > self.best_so_far_y):
            self.best_so_far_x, self.best_so_far_y = np.copy(x), -y
        return float(y)

    def _check_terminations(self):
        self.runtime = time.time() - self.start_time
        if self.n_function_evaluations >= self.max_function_evaluations:
            termination_signal = True, Terminations.MAX_FUNCTION_EVALUATIONS
        elif self.runtime >= self.max_runtime:
            termination_signal = True, Terminations.MAX_RUNTIME
        elif not self._is_maximization and (self.best_so_far_y <= self.fitness_threshold):
            termination_signal = True, Terminations.FITNESS_THRESHOLD
        elif self._is_maximization and (self.best_so_far_y >= self.fitness_threshold):
            termination_signal = True, Terminations.FITNESS_THRESHOLD
        else:
            termination_signal = False, Terminations.NO_TERMINATION
        self.termination_signal = termination_signal[1]
        return termination_signal[0]

    def _compress_fitness(self, fitness):
        fitness = np.array(fitness)
        # arrange in non-increasing order
        for i in range(len(fitness) - 1):
            if fitness[i] < fitness[i + 1]:
                fitness[i + 1] = fitness[i]
        # use 1-based index
        index = np.arange(1, len(fitness), self.record_fitness_frequency)
        # recover 0-based index via - 1
        index = np.append(index, len(fitness)) - 1
        self.fitness = np.stack((index, fitness[index]), 1)

    def _collect_results(self, fitness):
        if self._is_maximization:
            self.best_so_far_y *= -1
        if self.record_fitness:
            self._compress_fitness(fitness[:self.n_function_evaluations])
        return {'best_so_far_x': self.best_so_far_x,
                'best_so_far_y': self.best_so_far_y,
                'n_function_evaluations': self.n_function_evaluations,
                'runtime': time.time() - self.start_time,
                'termination_signal': self.termination_signal,
                'time_function_evaluations': self.time_function_evaluations,
                'fitness': self.fitness}

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def optimize(self, fitness_function=None):
        self.start_time = time.time()
        if fitness_function is not None:
            self.fitness_function = fitness_function
        fitness = []  # store all fitness generated during evolution
        return fitness



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CEM(Optimizer):
    """Cross-Entropy Method (CEM).

    References
    ----------
    Kroese, D.P., Porotsky, S. and Rubinstein, R.Y., 2006.
    The cross-entropy method for continuous multi-extremal optimization.
    Methodology and Computing in Applied Probability, 8(3), pp.383-407.
    https://link.springer.com/article/10.1007/s11009-006-9753-0

    De Boer, P.T., Kroese, D.P., Mannor, S. and Rubinstein, R.Y., 2005.
    A tutorial on the cross-entropy method.
    Annals of Operations Research, 134(1), pp.19-67.
    https://link.springer.com/article/10.1007/s10479-005-5724-z

    Rubinstein, R.Y. and Kroese, D.P., 2004.
    The cross-entropy method: a unified approach to combinatorial optimization,
        Monte-Carlo simulation, and machine learning.
    New York: Springer.
    https://link.springer.com/book/10.1007/978-1-4757-4321-0
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of individuals (samples) in each iteration
            self.n_individuals = 1000
        if self.n_parents is None:  # number of elitists for sampling distribution update
            self.n_parents = 200
        self.mean = options.get('mean')  # mean of Gaussian search (sampling/mutation) distribution
        if self.mean is None:  # 'mean' has a priority over 'x'
            self.mean = options.get('x')
        assert self.mean is not None
        self.sigma = options.get('sigma')  # global (overall) step-size
        assert self.sigma is not None
        self._sigmas = self.sigma*np.ones((self.ndim_problem,))  # individual step-sizes
        self._n_generations = 0

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _initialize_mean(self, is_restart=False):
        if is_restart or (self.mean is None):
            mean = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            mean = np.copy(self.mean)
        return mean

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['mean'] = mean
        results['_sigmas'] = self._sigmas
        results['_n_generations'] = self._n_generations
        return results



import numpy as np
import numba as nb

from pypop7.optimizers.es.es import ES


@nb.jit(nopython=True)
def cholesky_update(rm, z, downdate):
    # https://github.com/scipy/scipy/blob/d20f92fce9f1956bfa65365feeec39621a071932/
    #     scipy/linalg/_decomp_cholesky_update.py
    rm, z, alpha, beta = rm.T, z, np.empty_like(z), np.empty_like(z)
    alpha[-1], beta[-1] = 1., 1.
    sign = -1 if downdate else 1
    for r in range(len(z)):
        a = z[r] / rm[r, r]
        alpha[r] = alpha[r - 1] + sign * np.power(a, 2)
        beta[r] = np.sqrt(alpha[r])
        z[r + 1:] -= a * rm[r, r + 1:]
        rm[r, r:] *= beta[r] / beta[r - 1]
        rm[r, r + 1:] += sign * a / (beta[r] * beta[r - 1]) * z[r + 1:]
    return rm.T


class OPOA2015(ES):
    """(1+1)-Active-CMA-ES (OPOA2015).
    Krause, O. and Igel, C., 2015, January.
    A more efficient rank-one covariance matrix update for evolution strategies.
    In Proceedings of ACM Conference on Foundations of Genetic Algorithms (pp. 129-136).
    https://dl.acm.org/doi/abs/10.1145/2725494.2725496
    """
    def __init__(self, problem, options):
        options['n_individuals'] = 1  # mandatory setting for OPOA2015
        options['n_parents'] = 1  # mandatory setting for OPOA2015
        ES.__init__(self, problem, options)
        if self.eta_sigma is None:
            self.eta_sigma = 1 / (1 + self.ndim_problem / 2)
        self.p_ts = options.get('p_ts', 2 / 11)
        self.c_p = options.get('c_p', 1 / 12)
        self.c_c = options.get('c_c', 2 / (self.ndim_problem + 2))
        self.c_cov = options.get('c_cov', 2 / (np.power(self.ndim_problem, 2) + 6))
        self.p_t = options.get('p_t', 0.44)
        self.c_m = options.get('c_m', 0.4 / (np.power(self.ndim_problem, 1.6) + 1))
        self.k = options.get('k', 5)
        self._ancestors = []
        self._c_cf = 1 - self.c_cov + self.c_cov * self.c_c * (2 - self.c_c)

    def initialize(self, args=None, is_restart=False):
        mean = self._initialize_mean(is_restart)
        y = self._evaluate_fitness(mean, args)  # fitness
        cf = np.diag(np.ones(self.ndim_problem,))  # Cholesky factorization
        best_so_far_y, p_s = np.copy(y), self.p_ts
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        return mean, y, cf, best_so_far_y, p_s, p_c

    def _cholesky_update(self, cf=None, alpha=None, beta=None, v=None):  # triangular rank-one update
        assert self.ndim_problem == v.size
        if beta < 0:
            downdate, beta = True, -beta
        else:
            downdate = False
        return cholesky_update(np.sqrt(alpha) * cf, np.sqrt(beta) * v, downdate)

    def iterate(self, args=None, mean=None, cf=None, best_so_far_y=None, p_s=None, p_c=None):
        # sample and evaluate (only one) offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        cf_z = np.dot(cf, z)
        x = mean + self.sigma * cf_z
        y = self._evaluate_fitness(x, args)
        if y <= best_so_far_y:
            self._ancestors.append(y)
            mean, best_so_far_y = x, y
            p_s = (1 - self.c_p) * p_s + self.c_p
            is_better = True
        else:
            p_s *= 1 - self.c_p
            is_better = False
        self.sigma *= np.exp(self.eta_sigma * (p_s - self.p_ts) / (1 - self.p_ts))
        if p_s >= self.p_t:
            p_c *= 1 - self.c_c
            cf = self._cholesky_update(cf, self._c_cf, self.c_cov, p_c)
        elif is_better:
            p_c = (1 - self.c_c) * p_c + np.sqrt(self.c_c * (2 - self.c_c)) * cf_z
            cf = self._cholesky_update(cf, 1 - self.c_cov, self.c_cov, p_c)
        elif len(self._ancestors) >= self.k and y > self._ancestors[-self.k]:
            del self._ancestors[0]
            c_m = np.minimum(self.c_m, 1 / (2 * np.dot(z, z) - 1))
            cf = self._cholesky_update(cf, 1 + c_m, -c_m, cf_z)
        return mean, cf, best_so_far_y, p_s, p_c, y

    def restart_initialize(self, args=None, mean=None, y=None, cf=None,
                           best_so_far_y=None, p_s=None, p_c=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, cf, best_so_far_y, p_s, p_c = self.initialize(args, is_restart)
            fitness.append(y)
            self._fitness_list = [best_so_far_y]
            self._ancestors = []
        return mean, y, cf, best_so_far_y, p_s, p_c

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, cf, best_so_far_y, p_s, p_c = self.initialize(args)
        fitness.append(y)
        while True:
            mean, cf, best_so_far_y, p_s, p_c, y = self.iterate(
                args, mean, cf, best_so_far_y, p_s, p_c)
            if self.record_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, cf, best_so_far_y, p_s, p_c = self.restart_initialize(
                    args, mean, y, cf, best_so_far_y, p_s, p_c, fitness)
        results = self._collect_results(fitness, mean)
        return results



"""Repeat the following paper for `JADE`:
    Ahrari, A., Deb, K. and Preuss, M., 2017.
    Multimodal optimization by covariance matrix self-adaptation evolution strategy with repelling subpopulations.
    Evolutionary computation, 25(3), pp.439-471.
    https://doi.org/10.1162/evco_a_00182

    Very close performance can be obtained by our code. Therefore, we argue that
    the repeatability of `JADE` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.es.rscmsa import RSCMSA


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'fitness_threshold': 1e-5,
               'seed_rng': 0,  # undefined in the original paper
               }
    rscmsa = RSCMSA(problem, options)
    results = rscmsa.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 7.09e-05
    # vs (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'fitness_threshold': 1e-5,
               'seed_rng': 0,  # undefined in the original paper
               }
    rscmsa = RSCMSA(problem, options)
    results = rscmsa.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 22.80
    # vs (from the original paper)

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10 * np.ones((ndim_problem,)),
               'upper_boundary': 10 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'fitness_threshold': 1e-5,
               'seed_rng': 0,  # undefined in the original paper
               }
    rscmsa = RSCMSA(problem, options)
    results = rscmsa.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 64.68
    # vs (from the original paper)



import numpy as np

from pypop7.optimizers.es.es import ES


class VDCMA(ES):
    """Linear Covariance Matrix Adaptation (VDCMA, VD-CMA).

    Reference
    ---------
    Akimoto, Y., Auger, A. and Hansen, N., 2014, July.
    Comparison-based natural gradient optimization in high dimension.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 373-380). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598258

    See the official Python version from Akimoto:
    https://gist.github.com/youheiakimoto/08b95b52dfbf8832afc71dfff3aed6c8
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        # for faster adaptation of covariance matrix (c_1, c_mu)
        self.c_factor = options.get('c_factor', np.maximum((self.ndim_problem - 5) / 6, 0.5))
        self.c_c = None
        self.c_1 = None
        self.c_mu = None
        self.c_s = None
        self.d_s = None
        self._v_n = None
        self._v_2 = None
        self._v_ = None
        self._v_p2 = None

    def initialize(self, is_restart=False):
        self.c_c = self.options.get('c_c', (4 + self._mu_eff / self.ndim_problem) / (
                self.ndim_problem + 4 + 2 * self._mu_eff / self.ndim_problem))
        self.c_1 = self.options.get('c_1', self.c_factor * 2 / (
                np.power(self.ndim_problem + 1.3, 2) + self._mu_eff))
        self.c_mu = self.options.get('c_mu', np.minimum(1 - self.c_1, self.c_factor * 2 * (
                self._mu_eff - 2 + 1 / self._mu_eff) / (np.power(self.ndim_problem + 2, 2) + self._mu_eff)))
        self.c_s = self.options.get('c_s', 1 / (2 * np.sqrt(self.ndim_problem / self._mu_eff) + 1))
        self.d_s = self.options.get('d_s', 1 + self.c_s + 2 * np.maximum(0, np.sqrt(
            (self._mu_eff - 1) / (self.ndim_problem + 1)) - 1))
        d = np.ones((self.ndim_problem,))  # diagonal vector for sampling distribution
        # principal search direction (vector) for sampling distribution
        v = self.rng_optimization.standard_normal((self.ndim_problem,)) / np.sqrt(self.ndim_problem)
        p_s = np.zeros((self.ndim_problem,))  # evolution path for step-size adaptation (MCSA)
        p_c = np.zeros((self.ndim_problem,))  # evolution path for covariance matrix adaptation (CMA)
        self._v_n = np.linalg.norm(v)
        self._v_2 = np.power(self._v_n, 2)
        self._v_ = v / self._v_n
        self._v_p2 = np.power(self._v_, 2)
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        zz = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return d, v, p_s, p_c, z, zz, x, mean, y

    def iterate(self, d=None, z=None, zz=None, x=None, mean=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, zz, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            zz[k] = d * (z[k] + (np.sqrt(1 + self._v_2) - 1) * (np.dot(z[k], self._v_) * self._v_))
            x[k] = mean + self.sigma * zz[k]
            y[k] = self._evaluate_fitness(x[k], args)
        return z, zz, x, y

    def _p_q(self, zz, w=0):
        zz_v_ = np.dot(zz, self._v_)
        if isinstance(w, int) and w == 0:
            p = np.power(zz, 2) - self._v_2 / (1 + self._v_2) * (zz_v_ * (zz * self._v_)) - 1
            q = zz_v_ * zz - ((np.power(zz_v_, 2) + 1 + self._v_2) / 2) * self._v_
        else:
            p = np.dot(w, np.power(zz, 2) - self._v_2 / (1 + self._v_2) * (zz_v_ * (zz * self._v_).T).T - 1)
            q = np.dot(w, (zz_v_ * zz.T).T - np.outer((np.power(zz_v_, 2) + 1 + self._v_2) / 2, self._v_))
        return p, q

    def _update_distribution(self, d=None, v=None, p_s=None, p_c=None, zz=None, x=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        # update mean
        mean = np.dot(self._w, x[order])
        # MCSA
        z = np.dot(self._w, zz[order]) / d
        z += (1 / np.sqrt(1 + self._v_2) - 1) * np.dot(z, self._v_) * self._v_
        p_s = (1 - self.c_s) * p_s + np.sqrt(self.c_s * (2 - self.c_s) * self._mu_eff) * z
        p_s_2 = np.dot(p_s, p_s)
        self.sigma *= np.exp((np.sqrt(p_s_2) / self._e_chi - 1) * self.c_s / self.d_s)
        h_s = p_s_2 < (2 + 4 / (self.ndim_problem + 1)) * self.ndim_problem
        # update restricted covariance matrix (d, v)
        p_c = (1 - self.c_c) * p_c + h_s * np.sqrt(
            self.c_c * (2 - self.c_c) * self._mu_eff) * np.dot(self._w, zz[order])

        gamma = 1 / np.sqrt(1 + self._v_2)
        alpha = np.sqrt(np.power(self._v_2, 2) + (1 + self._v_2) / np.max(
            self._v_p2) * (2 - gamma)) / (2 + self._v_2)
        if alpha < 1:
            beta = (4 - (2 - gamma) / np.max(self._v_p2)) / np.power(1 + 2 / self._v_2, 2)
        else:
            alpha, beta = 1, 0
        b = 2 * np.power(alpha, 2) - beta
        a = 2 - (b + 2 * np.power(alpha, 2)) * self._v_p2
        _v_p2_a = self._v_p2 / a

        if self.c_mu == 0:
            p_mu, q_mu = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,))
        else:
            p_mu, q_mu = self._p_q(zz[order] / d, self._w)
        if self.c_1 == 0:
            p_1, q_1 = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,))
        else:
            p_1, q_1 = self._p_q(p_c / d)
        p = self.c_mu * p_mu + h_s * self.c_1 * p_1
        q = self.c_mu * q_mu + h_s * self.c_1 * q_1

        if self.c_mu + self.c_1 > 0:
            r = p - alpha / (1 + self._v_2) * ((2 + self._v_2) * (
                    q * self._v_) - self._v_2 * np.dot(self._v_, q) * self._v_p2)
            s = r / a - b * np.dot(r, _v_p2_a) / (1 + b * np.dot(self._v_p2, _v_p2_a)) * _v_p2_a
            ng_v = q / self._v_n - alpha / self._v_n * ((2 + self._v_2) * (
                    self._v_ * s) - np.dot(s, np.power(self._v_, 2)) * self._v_)
            ng_d = d * s
            up_factor = np.minimum(1, 0.7 * self._v_n / np.sqrt(np.dot(ng_v, ng_v)))
            up_factor = np.minimum(up_factor, 0.7 * (d / np.min(np.abs(ng_d))))
        else:
            ng_v, ng_d, up_factor = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,)), 1
        v += up_factor * ng_v
        d += up_factor * ng_d
        self._v_n = np.linalg.norm(v)
        self._v_2 = np.power(self._v_n, 2)
        self._v_ = v / self._v_n
        self._v_p2 = np.power(self._v_, 2)
        return mean, p_s, p_c, v, d

    def restart_initialize(self, d=None, v=None, p_s=None, p_c=None, z=None, zz=None, x=None, mean=None, y=None):
        if ES.restart_initialize(self):
            d, v, p_s, p_c, z, zz, x, mean, y = self.initialize(True)
        return d, v, p_s, p_c, z, zz, x, mean, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        d, v, p_s, p_c, z, zz, x, mean, y = self.initialize()
        while True:
            z, zz, x, y = self.iterate(d, z, zz, x, mean, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p_s, p_c, v, d = self._update_distribution(d, v, p_s, p_c, zz, x, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                d, v, p_s, p_c, z, zz, x, mean, y = self.restart_initialize(d, v, p_s, p_c, z, zz, x, mean, y)
        results = self._collect_results(fitness, mean)
        results['d'] = d
        results['v'] = v
        return results



from pypop7.optimizers.de.de import DE
from pypop7.optimizers.de.cde import CDE
from pypop7.optimizers.de.tde import TDE
from pypop7.optimizers.de.jade import JADE
from pypop7.optimizers.de.code import CODE


__all__ = [DE, CDE, TDE, JADE, CODE]



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class DE(Optimizer):
    """Differential Evolution (DE).

    This is the **base** (abstract) class for all `DE` classes. Please use any of its instantiated subclasses
    to optimize the black-box problem at hand.

    .. note:: Originally `DE` was proposed to solve some challenging real-world problems by Kenneth Price and
       Rainer Storn, `recipients of Evolutionary Computation Pioneer Award 2017 <https://tinyurl.com/456as566>`_.
       Although there are *few* significant theoretical advances (to our knowledge), it is **still widely used in
       practice** owing to its often attractive performance on many multimodal black-box functions.

       The popular and powerful `SciPy <https://www.nature.com/articles/s41592-019-0686-2>`_ has provided an
       open-source implementation for `DE`.

       *"DE borrows the idea from Nelder&Mead of employing information from within the vector population to alter
       the search space."* --- `Storn&Price, 1997, JGO <https://doi.org/10.1023/A:1008202821328>`_

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose information during optimization (`bool`, default:
                  `True`),
                * 'verbose_frequency'        - generation frequency of printing verbose information (`int`, default:
                  `10`);
              and with one particular setting (`key`):
                * 'n_individuals' - number of offspring, offspring population size (`int`, default: `100`).

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, offspring population size. For `DE`, typically a *large* (often >=100)
                    population size is used to better explore for high-dimensional multimodal functions. Obviously
                    the *optimal* population size is problem-dependent, relying on fine-tuning.

    Methods
    -------

    References
    ----------
    Price, K.V., 2013.
    Differential evolution.
    In Handbook of Optimization (pp. 187-214). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-30504-7_8

    Price, K.V., Storn, R.M. and Lampinen, J.A., 2005.
    Differential evolution: A practical approach to global optimization.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/3-540-31306-0

    Storn, R.M. and Price, K.V. 1997.
    Differential evolution – a simple and efficient heuristic for global optimization over continuous spaces.
    Journal of Global Optimization, 11(4), pp.341–359.
    https://doi.org/10.1023/A:1008202821328
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of offspring, offspring population size
            self.n_individuals = 100
        self._n_generations = 0  # number of generations

    def initialize(self):
        raise NotImplementedError

    def mutate(self):
        raise NotImplementedError

    def crossover(self):
        raise NotImplementedError

    def select(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness=None):
        results = Optimizer._collect_results(self, fitness)
        results['_n_generations'] = self._n_generations
        return results



import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.r1es import R1ES


class RMES(R1ES):
    """Rank-M Evolution Strategy (RMES).

    Reference
    ---------
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
    """
    def __init__(self, problem, options):
        R1ES.__init__(self, problem, options)
        self.n_evolution_paths = options.get('n_evolution_paths', 2)  # m in Algorithm 2
        self.generation_gap = options.get('generation_gap', self.ndim_problem)  # T in Algorithm 2
        self._a = np.sqrt(1 - self.c_cov)  # for Line 4 in Algorithm 3
        self._a_m = np.power(self._a, self.n_evolution_paths)  # for Line 4 in Algorithm 3
        self._b = np.sqrt(self.c_cov)  # for Line 4 in Algorithm 3

    def initialize(self, args=None, is_restart=False):
        x, mean, p, s, y = R1ES.initialize(self, args, is_restart)
        mp = np.zeros((self.n_evolution_paths, self.ndim_problem))  # multiple evolution paths
        t_hat = np.zeros((self.n_evolution_paths,))  # for Algorithm 2
        return x, mean, p, s, mp, t_hat, y

    def iterate(self, x=None, mean=None, mp=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))
            sum_p = np.zeros((self.ndim_problem,))
            for i in np.arange(self.n_evolution_paths) + 1:
                r = self.rng_optimization.standard_normal()
                sum_p += np.power(self._a, self.n_evolution_paths - i) * r * mp[i - 1]
            x[k] = mean + self.sigma * (self._a_m * z + self._b * sum_p)
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, p=None, s=None,
                             mp=None, t_hat=None, y=None, y_bak=None):
        mean, p, s = R1ES._update_distribution(self, x, mean, p, s, y, y_bak)
        # update multiple evolution paths
        t_min = np.min(np.diff(t_hat))  # Line 2 in Algorithm 2 (T_min)
        if (t_min > self.generation_gap) or (self._n_generations < self.n_evolution_paths):
            for i in range(self.n_evolution_paths - 1):
                mp[i], t_hat[i] = mp[i + 1], t_hat[i + 1]
        else:
            i_apostrophe = np.argmin(np.diff(t_hat))  # Line 6 in Algorithm 2 (i')
            for i in range(i_apostrophe, self.n_evolution_paths - 1):
                mp[i], t_hat[i] = mp[i + 1], t_hat[i + 1]
        mp[-1], t_hat[-1] = p, self._n_generations
        return mean, p, s, mp, t_hat

    def restart_initialize(self, args=None, x=None, mean=None, p=None, s=None,
                           mp=None, t_hat=None, y=None, fitness=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            x, mean, p, s, mp, t_hat, y = self.initialize(args, is_restart)
            fitness.append(y[0])
            self.d_sigma *= 2
        return x, mean, p, s, mp, t_hat, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, s, mp, t_hat, y = self.initialize(args)
        fitness.append(y[0])
        while True:
            y_bak = np.copy(y)
            x, y = self.iterate(x, mean, mp, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, s, mp, t_hat = self._update_distribution(x, mean, p, s, mp, t_hat, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, p, s, mp, t_hat, y = self.restart_initialize(
                    args, x, mean, p, s, mp, t_hat, y, fitness)
        results = self._collect_results(fitness, mean)
        results['p'] = p
        results['s'] = s
        return results



import numpy as np

from pypop7.optimizers.cem.cem import CEM


class SCEM(CEM):
    """Standard Cross-Entropy Method (SCEM).

    References
    ----------
    Kroese, D.P., Porotsky, S. and Rubinstein, R.Y., 2006.
    The cross-entropy method for continuous multi-extremal optimization.
    Methodology and Computing in Applied Probability, 8(3), pp.383-407.
    https://link.springer.com/article/10.1007/s11009-006-9753-0
    (See [B Main CE Program] for the official Matlab code.)

    De Boer, P.T., Kroese, D.P., Mannor, S. and Rubinstein, R.Y., 2005.
    A tutorial on the cross-entropy method.
    Annals of Operations Research, 134(1), pp.19-67.
    https://link.springer.com/article/10.1007/s10479-005-5724-z
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)
        self.alpha = options.get('alpha', 0.8)  # smoothing factor

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)
        x = np.empty((self.n_individuals, self.ndim_problem))  # samples (population)
        y = np.empty((self.n_individuals,))  # fitness (cost)
        return mean, x, y

    def iterate(self, mean=None, x=None, y=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            x[i] = mean + self._sigmas*self.rng_optimization.standard_normal((self.ndim_problem,))
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def _update_parameters(self, mean=None, x=None, y=None):
        xx = x[np.argsort(y)[:self.n_parents]]
        mean = self.alpha*np.mean(xx, axis=0) + (1.0-self.alpha)*mean
        self._sigmas = self.alpha*np.std(xx, axis=0) + (1.0-self.alpha)*self._sigmas
        return mean

    def optimize(self, fitness_function=None, args=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, x, y = self.initialize()
        while True:
            x, y = self.iterate(mean, x, y, args)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            mean = self._update_parameters(mean, x, y)
        return self._collect_results(fitness, mean)



"""Latent Action Monto Carlo Tree Search(LA-MCTS)
    Reference
    --------------
    L. Wang, R. Fonseca, Y. Tian
    Learning Search Space Partition for Black-box Optimization using Monte Carlo Tree Search
    NeurIPS 2020
    https://proceedings.neurips.cc/paper/2020/hash/e2ce14e81dba66dbff9cbc35ecfdb704-Abstract.html
    Code of paper:
    https://github.com/facebookresearch/LaMCTS/tree/main/LA-MCTS
"""
import numpy as np
import torch
from sklearn.cluster import KMeans
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import ConstantKernel, Matern
from scipy.stats import norm
import copy as cp
from torch.quasirandom import SobolEngine

from pypop7.optimizers.core.optimizer import Optimizer


# generate initial points
def latin_hyper_cube(n, dims, ub, lb):
    points = np.zeros((n, dims))
    centers = (1.0 + 2.0 * np.arange(0.0, n))
    centers = centers / float(2 * n)
    for i in range(0, dims):
        points[:, i] = centers[np.random.permutation(n)]
    perturbation = np.random.uniform(-1.0, 1.0, (n, dims))
    perturbation = perturbation / float(2 * n)
    points += perturbation
    points = points * (ub - lb) + lb
    return points


# Use for classifying points
class Classifier:
    def __init__(self, samples, dimension, kernel_type, gamma_type):
        self.dimension = dimension
        # create gaussian regressor
        noise = 0.1
        m52 = ConstantKernel(1.0) * Matern(length_scale=1.0, nu=2.5)
        self.gpr = GaussianProcessRegressor(kernel=m52, alpha=noise**2)
        self.kmeans = KMeans(n_clusters=2)
        # data structure
        self.samples = []
        self.x = []
        self.fx = []
        # learn boundary
        self.kernel_type = kernel_type
        self.gamma_type = gamma_type
        self.svm = SVC(kernel=kernel_type, gamma=gamma_type)
        self.update_samples(samples)

    def get_mean(self):
        if len(self.fx) == 0:
            return float('inf')
        else:
            return np.mean(self.fx)

    def update_samples(self, samples):
        self.x = []
        self.fx = []
        self.samples = samples
        for sample in samples:
            self.x.append(sample[0])
            self.fx.append(sample[1])
        self.x = np.asarray(self.x, dtype=np.float64).reshape(-1, self.dimension)
        self.fx = np.asarray(self.fx, dtype=np.float64).reshape(-1)

    def learn_cluster(self):
        data = np.concatenate((self.x, self.fx.reshape([-1, 1])), axis=1)
        self.kmeans = self.kmeans.fit(data)
        plabel = self.kmeans.predict(data)
        zero_label_fx = []
        one_label_fx = []
        for i in range(len(plabel)):
            if plabel[i] == 0:
                zero_label_fx.append(self.fx[i])
            elif plabel[i] == 1:
                one_label_fx.append(self.fx[i])
            else:
                print("There should only be two cluster")
        zero_label_mean = np.mean(zero_label_fx)
        one_label_mean = np.mean(one_label_fx)
        if zero_label_mean > one_label_mean:
            for i in range(len(plabel)):
                if plabel[i] == 0:
                    plabel[1] = 1
                else:
                    plabel[i] = 0
        return plabel

    def expected_improvement(self, x, xi=0.001):
        x_sample = self.x
        mu, std = self.gpr.predict(x, return_std=True)
        mu_sample = self.gpr.predict(x_sample)
        y_max = np.max(mu_sample)
        a = (mu - y_max - xi)
        z = a / std
        return a * norm.cdf(z) + std * norm.pdf(z)

    def get_sample_ratio_in_region(self, x, path):
        length = len(x)
        for node in path:
            if len(x) == 0:
                return 0, np.array([])
            boundary = node[0].classifier.svm
            x = x[boundary.predict(x) == node[1]]
        ratio = len(x) / length
        return ratio, x

    def split_data(self):
        good_samples = []
        bad_samples = []
        if len(self.samples) == 0:
            return good_samples, bad_samples
        plabel = self.learn_cluster()
        assert len(plabel) == len(self.x)
        if (1 in plabel) == True:
            self.svm.fit(self.x, plabel)
            for i in range(len(plabel)):
                if plabel[i] == 0:
                    good_samples.append(self.samples[i])
                else:
                    bad_samples.append((self.samples[i]))
            assert len(good_samples) + len(bad_samples) == len(self.samples)
            return good_samples, bad_samples, True
        else:
            return self.samples, [], False

    def propose_rand_sample_sobol(self, sample_num, path, lower_boundary, upper_boundary):
        seed = np.random.randint(int(1e6))
        sobol = SobolEngine(dimension=self.dimension, scramble=True, seed=seed)

        ratio_check, centers = self.get_sample_ratio_in_region(self.x, path)
        if ratio_check == 0 or len(centers) == 0:
            return np.random.uniform(lower_boundary, upper_boundary, size=(sample_num, self.dimension))

        final_cands = []
        for center in centers:
            center = self.x[np.random.randint(len(self.x))]
            cands = sobol.draw(2000).to(dtype=torch.float64).cpu().detach().numpy()
            ratio = 1
            leng = 0.0001
            Blimit = np.max(upper_boundary - lower_boundary)
            while ratio == 1 and leng < Blimit:
                lb = np.clip(center - leng/2, lower_boundary, upper_boundary)
                ub = np.clip(center + leng/2, lower_boundary, upper_boundary)
                cands_ = cp.deepcopy(cands)
                cands_ = (ub - lb) * cands_ + lb
                ratio, cands_ = self.get_sample_ratio_in_region(cands_, path)
                if ratio < 1:
                    final_cands.extend(cands_.tolist())
                leng *= 2
        final_cands = np.array(final_cands)
        if len(final_cands) > sample_num:
            final_cands_idx = np.random.choice(len(final_cands), sample_num)
            return final_cands[final_cands_idx]
        else:
            if len(final_cands) == 0:
                return np.random.uniform(lower_boundary, upper_boundary, size=(sample_num, self.dimension))
            else:
                return final_cands

    def propose_sample_bo(self, sample_num, path, lower_bound, upper_bound, samples):
        # train the gpr
        x = []
        fx = []
        for sample in samples:
            x.append(sample[0])
            fx.append(sample[1])
        x = np.asarray(x).reshape(-1, self.dimension)
        fx = np.asarray(fx).reshape(-1)
        self.gpr.fit(x, fx)

        if len(path) == 0:
            return np.random.uniform(lower_bound, upper_bound, size=(sample_num, self.dimension))
        nums_rand_samples = 10000
        x = self.propose_rand_sample_sobol(nums_rand_samples, path, lower_bound, upper_bound)
        if len(x) == 0:
            return np.random.uniform(lower_bound, upper_bound, size=(sample_num, self.dimension))
        # select x with least expect improvement
        expect_improves = self.expected_improvement(x, xi=0.001)
        order = np.argsort(expect_improves)[:sample_num]
        return x[order]


# Use for build tree
class Node:
    obj_counter = 0
    # If a leave holds >= SPLIT_THRESH, we split into two new nodes.

    def __init__(self, parent=None, dimension=10, is_reset=False, kernel_type='rbf', gamma_type='auto'):
        self.dimension = dimension
        self.parent = parent
        self.kids = []  # 0:good, 1:bad
        self.n = 0  # represent number of visit this node
        self.value = float('inf')  # node value equal to mean of object values in objects set
        self.objects = []
        self.is_svm_splittable = False
        self.classifier = Classifier([], self.dimension, kernel_type, gamma_type)
        if is_reset:
            Node.obj_counter = 0
        self.id = Node.obj_counter
        Node.obj_counter += 1

    def update_kids(self, good_kid, bad_kid):
        assert len(self.kids) == 0
        self.kids.append(good_kid)
        self.kids.append(bad_kid)
        assert self.kids[0].classifier.get_mean() < self.kids[1].classifier.get_mean()

    def is_leaf(self):
        if len(self.kids) == 0:
            return True
        else:
            return False

    def train_and_split(self):
        self.classifier.update_samples(self.objects)
        good_kid_data, bad_kid_data, is_split = self.classifier.split_data()
        return good_kid_data, bad_kid_data, is_split

    def update_objects(self, samples):
        self.objects.clear()
        self.objects.extend(samples)
        self.classifier.update_samples(self.objects)
        if len(self.objects) <= 2:
            self.is_svm_splittable = False
        else:
            self.is_svm_splittable = True
        self.value = self.classifier.get_mean()
        self.n = len(self.objects)

    def propose_sample_bo(self, sample_num, path, lower_boundary, upper_boundary, samples):
        propose_x = self.classifier.propose_sample_bo(sample_num, path, lower_boundary,
                                                      upper_boundary, samples)
        return propose_x

    def clear_data(self):
        self.objects.clear()

    def visit(self):
        self.n += 1

    def get_ucb(self, cp):
        if self.n == 0 or self.parent is None:
            return float('inf')
        else:
            # according to the paper
            # return self.value / self.n + 2 * Cp * np.sqrt(2 * np.log(self.parent.n) / self.n)
            # according to the code
            return self.value + 2 * cp * np.sqrt(2 * np .power(self.parent.n, 0.5) / self.n)

    def get_name(self):
        return "node" + str(self.id)

    def get_value(self):
        return self.value


class LAMCTS(Optimizer):
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.kernel_type = options.get('kernel_type')
        self.gamma_type = options.get('gamma_type')
        self.Cp = options.get('Cp')
        self.leaf_size = options.get('leaf_size')
        self.solver_type = options.get('solver_type')
        self._n_generations = 0
        self.nodes = []
        self.samples = []
        self.sample_counter = 0
        self.current_best_sample = None
        self.current_best_value = float('inf')

        # initialize the root
        root = Node(parent=None, dimension=self.ndim_problem,
                    is_reset=True, kernel_type=self.kernel_type, gamma_type=self.gamma_type)
        self.nodes.append(root)
        self.ROOT = root

    def initialize(self, is_restart=False):
        values = []
        points = latin_hyper_cube(self.n_individuals, self.ndim_problem,
                                  self.upper_boundary, self.lower_boundary)
        for point in points:
            values.append(self.collect_samples(point))
        return values

    def get_leaf_status(self):
        status = []
        for node in self.nodes:
            if node.is_leaf() is True and len(node.objects) > self.leaf_size and node.is_svm_splittable is True:
                status.append(True)
            else:
                status.append(False)
        return np.array(status)

    def dynamic_treeify(self):
        # clean the original tree
        self.ROOT.obj_counter = 0
        for node in self.nodes:
            node.clear_data()
        self.nodes.clear()

        # reset the root
        new_root = Node(parent=None, dimension=self.ndim_problem, is_reset=True,
                        kernel_type=self.kernel_type, gamma_type=self.gamma_type)
        self.nodes.append(new_root)
        self.ROOT = new_root
        self.ROOT.update_objects(self.samples)

        # build the tree
        while True:
            status = self.get_leaf_status()
            if True in status:
                split_by_samples = np.argwhere(self.get_leaf_status() == True).reshape(-1)
                for idx in split_by_samples:
                    parent = self.nodes[idx]
                    good_kid_data, bad_kid_data, is_split = parent.train_and_split()
                    if is_split is True:
                        good_kid = Node(parent, self.ndim_problem, False, self.kernel_type, self.gamma_type)
                        bad_kid = Node(parent, self.ndim_problem, False, self.kernel_type, self.gamma_type)
                        good_kid.update_objects(good_kid_data)
                        bad_kid.update_objects(bad_kid_data)
                        if good_kid.classifier.get_mean() < bad_kid.classifier.get_mean():
                            parent.update_kids(good_kid, bad_kid)
                            self.nodes.append(good_kid)
                            self.nodes.append(bad_kid)
                        else:
                            parent.update_kids(bad_kid, good_kid)
                            self.nodes.append(bad_kid)
                            self.nodes.append(good_kid)
                    else:
                        self.nodes[idx].is_svm_splittable = False
            else:
                break

    def collect_samples(self, x, value=None):
        if value is None:
            value = self._evaluate_fitness(x)
        if value < self.current_best_value:
            self.current_best_value = value
            self.current_best_sample = x
        self.sample_counter += 1
        self.samples.append((x, value))
        return value

    def select(self):
        current_node = self.ROOT
        path = []
        while current_node.is_leaf() is False:
            uct = []
            for kid in current_node.kids:
                uct.append(kid.get_ucb(self.Cp))
            choice = np.random.choice(np.argwhere(uct == np.amin(uct)).reshape(-1), 1)[0]
            path.append((current_node, choice))
            current_node = current_node.kids[choice]
        return current_node, path

    # refresh the value and num of a line of tree
    def back_propagate(self, leaf, value):
        current_node = leaf
        while current_node is not None:
            current_node.value = (current_node.value * current_node.n + value) / (current_node.value + 1)
            current_node.n += 1
            current_node = current_node.parent

    def iterate(self, leaf, path):
        values = []
        if self.solver_type == 'bo':
            samples = leaf.propose_sample_bo(1, path, self.lower_boundary,
                                             self.upper_boundary, self.samples)
        else:
            raise Exception("Solver not implemented")
        for i in range(len(samples)):
            if self.solver_type == 'bo':
                value = self.collect_samples(samples[i])
            else:
                raise Exception("Solver not implemented")
            self.back_propagate(leaf, value)
            values.append(value)
        return values

    def optimize(self, fitness_function=None):
        fitness = Optimizer.optimize(self, fitness_function)
        values = self.initialize()
        if self.record_fitness:
            fitness.extend(values)
        while True:
            self.dynamic_treeify()
            leaf, path = self.select()
            values = self.iterate(leaf, path)
            if self.record_fitness:
                fitness.extend(values)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(values)
        results = self._collect_results(fitness)
        return results

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            best_so_far_y = -self.best_so_far_y if self._is_maximization else self.best_so_far_y
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness):
        results = Optimizer._collect_results(self, fitness)
        results['_n_generations'] = self._n_generations
        return results



from pypop7.optimizers.rs.rs import RS
from pypop7.optimizers.rs.prs import PRS
from pypop7.optimizers.rs.rhc import RHC
from pypop7.optimizers.rs.arhc import ARHC
from pypop7.optimizers.rs.srs import SRS


__all__ = [RS, PRS, RHC, ARHC, SRS]



import numpy as np
from scipy.stats import norm

from pypop7.optimizers.es.es import ES


def mahalanobis_distance(y, x, inv_cov):
    if y.ndim == 1:
        return np.sqrt(np.dot(np.dot(y - x, inv_cov), (y - x)))
    if y.ndim == 2:
        return np.array([np.sqrt(np.dot(np.dot(yy - x, inv_cov), (yy - x))) for yy in y])


class RSCMSA(ES):
    """Covariance Matrix Self-Adaptation Evolution Strategy with Repelling Subpopulations (RS-CMSA)

    Reference
    ---------
    Beyer, H.G. and Sendhoff, B., 2008.
    Covariance matrix adaptation revisited–the CMSA evolution strategy.
    In International Conference on Parallel Problem-Solving from Nature (pp. 123-132). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-540-87700-4_13

    Ahrari, A., Deb, K. and Preuss, M., 2017.
    Multimodal optimization by covariance matrix self-adaptation evolution strategy with repelling subpopulations.
    Evolutionary computation, 25(3), pp.439-471.
    https://doi.org/10.1162/evco_a_00182
    """

    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.tau = np.sqrt(0.5 / self.ndim_problem)  # the learning rate of global step size
        self.c_cov = self.n_parents / (self.n_parents + self.ndim_problem * (self.ndim_problem + 1))  # 1/tau_c
        self.n_elt = np.maximum(1, int(0.15 * self.n_individuals))  # number of elites
        self.tau_hat_d = np.sqrt(1 / self.ndim_problem)  # the learning rate of the normalized taboo distance
        self.c_red = np.power(0.99, 1 / self.ndim_problem)  # the size of each taboo region shrinks by (1-self.c_red)
        self.alpha_new = options.get('alpha_new', 0.5)  # target rate of basin identification
        self.c_threshold = options.get('c_threshold', 0.01)  # criticality threshold for taboo points
        self.n_s = options.get('n_subpopulations', 10)  # number of subpopulations
        self.p = options.get('p', 25)  # percentile
        self.budget = options.get('budget', 10)  # max evaluation budget for the hill-valley function
        self.hat_d_0 = options.get('hat_d_0', 1)  # default value of the normalized taboo distance
        self.n_condition = options.get('n_condition', 1e14)  # condition number
        self.sigma = options.get('sigma', 0.25)
        self._n_s_bak = np.copy(self.n_s)
        self._best_so_far_y_bak = np.copy(self.best_so_far_y)

    def initialize(self, args=None):
        archive_x, archive_f, archive_d = np.empty((0, self.ndim_problem)), np.empty((0,)), np.empty((0,))
        return archive_x, archive_f, archive_d

    def is_new_basin(self, x1, f1, x4, f4, args=None):
        new_basin, n = False, 0
        if np.sqrt(np.sum(np.power(x1 - x4, 2))) > 0:
            max_f = np.maximum(f1, f4)
            direction = x4 - x1  # search direction
            delta = np.sqrt(np.sum(np.power(direction, 2)))
            direction = direction / delta / 2.618
            x2 = x1 + delta * direction
            f2 = self._evaluate_fitness(x2, args)
            n += 1
            if f2 > max_f:
                new_basin = True
            elif n < self.budget:  # Max evaluation budget for the hill-valley ( Detect Multimodal) function
                x3 = x1 + 1.618 * delta * direction
                f3 = self._evaluate_fitness(x3, args)
                n += 1
                if f3 > max_f:
                    new_basin = True
                else:
                    while (n < self.budget) and (np.sqrt(np.sum(np.power(x1 - x4, 2))) > 0):
                        if f2 < f3:
                            delta = np.sqrt(np.sum(np.power(x4 - x2, 2))) / 2.618
                            x1, f1 = np.copy(x2), np.copy(f2)
                            x2, f2 = np.copy(x3), np.copy(f3)
                            x3 = x1 + 1.618 * delta * direction
                            f3 = self._evaluate_fitness(x3, args)
                            n += 1
                        elif f2 > f3:
                            delta = np.sqrt(np.sum(np.power(x3 - x1, 2))) / 2.618
                            x4, f4 = np.copy(x3), np.copy(f3)
                            x3, f3 = np.copy(x2), np.copy(f2)
                            x2 = x1 + delta * direction
                            f2 = self._evaluate_fitness(x2, args)
                            n += 1
                        else:
                            delta = np.sqrt(np.sum(np.power(x3 - x2, 2))) / 2.618034
                            x1, f1 = np.copy(x2), np.copy(f2)
                            x4, f4 = np.copy(x3), np.copy(f3)
                            x2 = x1 + delta * direction
                            x3 = x4 - delta * direction
                            f2 = self._evaluate_fitness(x2, args)
                            f3 = self._evaluate_fitness(x3, args)
                            n += 2
                        if np.maximum(f2, f3) > max_f:
                            new_basin = True
                            break
        return new_basin

    def update_population_size(self, average_used_iteration_number):
        factor = (self.max_function_evaluations - self.n_function_evaluations) / (
                self.n_individuals * self._n_s_bak * average_used_iteration_number)
        if factor >= 2:
            self.n_individuals = 2 * self.n_individuals
            self.n_s = np.copy(self._n_s_bak)
        elif factor >= 1:
            self.n_individuals = int(self.n_individuals ** 2 * factor)
            self.n_s = np.copy(self._n_s_bak)
        else:
            self.n_s = np.max([1, int(self._n_s_bak * factor)])
        self.n_parents = int(self.n_individuals / 2)
        w_base, w = np.log((self.n_individuals + 1) / 2), np.log(np.arange(self.n_parents) + 1)
        self._w = (w_base - w) / (self.n_parents * w_base - np.sum(w))
        self._mu_eff = 1 / np.sum(np.power(self._w, 2))
        self.c_cov = self.n_parents / (self.n_parents + self.ndim_problem * (self.ndim_problem + 1))  # 1/tau_c
        self.n_elt = np.maximum(1, int(0.15 * self.n_individuals))  # number of elites

    def update_archive(self, best_x, best_y, archive_x, archive_y, archive_d, args=None):
        hat_d = self.hat_d_0 if len(archive_d) == 0 else np.percentile(archive_d, self.p)
        if len(archive_y) > 0:
            if (np.min(best_y) + self.fitness_threshold) < np.min(archive_y):  # discard archive solutions
                archive_x, archive_y, archive_d = np.empty((0, self.ndim_problem)), np.empty((0,)), np.empty((0,))

        # Consider only global minima, among the recently generated solutions
        index = np.arange(self.n_s)[best_y <= (self.fitness_threshold + np.min(np.hstack((archive_y, best_y))))]
        n_rep = np.zeros((len(archive_x),))  # number of subpopulations that have converged to corresponding basin
        for i in index:
            archive_x = archive_x[np.argsort([np.sqrt(np.sum(np.power(best_x[i] - x, 2))) for x in archive_x])]  # sort
            is_new = 1
            for j in range(len(archive_x)):
                is_basin = self.is_new_basin(best_x[i], best_y[i], archive_x[j], archive_y[j], args)
                if self._check_terminations():
                    return archive_x, archive_y, archive_d

                if is_basin:  # share the same basin
                    if best_y[i] < archive_y[j]:
                        archive_x[j], archive_y[j] = best_x[i], best_y[i]
                    n_rep[j] += 1
                    is_new = 0  # not a new basin
                    break
            if is_new == 1:
                archive_x = np.vstack((archive_x, best_x[i]))
                archive_y = np.hstack((archive_y, best_y[i]))
                archive_d = np.hstack((archive_d, hat_d))
                n_rep = np.hstack((n_rep, 0))

        diff = n_rep - self.alpha_new * (len(index) / len(archive_x))
        for i in range(len(archive_x)):
            if diff[i] > 0:
                archive_d[i] *= (1 + diff[i]) ** self.tau_hat_d
            else:
                archive_d[i] *= (1 - diff[i]) ** (-1 * self.tau_hat_d)
        return archive_x, archive_y, archive_d

    def initialize_subpopulation(self, archive_xx, archive_d):
        hat_d = self.hat_d_0 if len(archive_d) == 0 else np.percentile(archive_d, self.p)
        base = (self.initial_upper_boundary - self.initial_lower_boundary) ** 2
        cov, inv_cov, sigma = np.diag(base), np.diag(1 / base), self.sigma
        means = np.empty((self.n_s, self.ndim_problem))
        for i in range(self.n_s):
            n_rej = 0
            while True:
                accept = True
                x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
                for j in range(len(archive_xx)):  # for archived points
                    accept = mahalanobis_distance(archive_xx[j], x, inv_cov) >= ((2 * hat_d + archive_d[j]) * sigma)
                    if not accept:
                        break
                if accept:
                    for k in range(i):  # for previously sampled points
                        accept = mahalanobis_distance(means[k], x, inv_cov) >= (3 * hat_d * sigma)
                        if not accept:
                            break
                if accept:
                    means[i] = x
                    break
                else:
                    n_rej += 1
                if n_rej > 100:
                    sigma, n_rej = self.c_red * sigma, 0  # shrink the size of each taboo region
        sigmas = sigma * np.ones((self.n_s,))
        means_d = hat_d * np.ones((self.n_s,))
        return means, sigmas, cov, means_d

    def iterate_subpopulation(self, mean, cov, sigma, elt_x, elt_y, elt_z, elt_s,
                              best_y_ne, median_y_ne, superior_mean, superior_d,
                              archive_x, archive_y, archive_d, args):
        # Singular value decomposition
        u, s, v = np.linalg.svd(cov)
        inv_cov = v @ np.diag(1 / s) @ u  # Inverse of the matrix
        sqrt_w = np.sqrt(s)
        sqrt_cov = u @ np.diag(sqrt_w)

        # Generate taboo points
        taboo, taboo_d = self.generate_taboo_point(superior_mean, superior_d, archive_x, archive_y, archive_d, elt_y)

        # Determine critical taboo points, Equation (12)
        c_index = self.index_critical_points(mean, sigma, np.max(sqrt_w), taboo, taboo_d)

        # Sample
        x, y, z, sigmas, scale = self.sample(mean, sigma, sqrt_cov, inv_cov, taboo, taboo_d, c_index, args)

        # for the non-elite solutions
        best_y_ne = np.hstack((best_y_ne, np.min(y)))
        median_y_ne = np.hstack((median_y_ne, np.median(y)))

        x, y, z, sigmas = self.recombine(
            inv_cov, sigma, x, y, z, sigmas, scale, taboo, taboo_d, c_index, elt_x, elt_y, elt_z, elt_s)

        # Equation (b)
        mean, cov, sigma = self._update_distribution(x, z, y, cov, sigma, sigmas)

        order = np.argsort(y)
        best_x, best_y = x[order[0]], y[order[0]]
        base = order[:self.n_elt]  # Update the elite solutions
        elt_x, elt_y, elt_z, elt_s = x[base], y[base], z[base], sigmas[base]
        return best_x, best_y, mean, cov, sigma, elt_x, elt_y, elt_z, elt_s, best_y_ne, median_y_ne

    def recombine(self, inv_cov, sigma, x, y, z, sigmas, scale, taboo, taboo_d, c_index, elt_x, elt_y, elt_z, elt_s):
        # Append the surviving elites from the previous generation
        for k in range(len(elt_y)):
            if elt_y[k] < self._best_so_far_y_bak:
                accept = False
                for c in c_index:
                    d = mahalanobis_distance(elt_x[k], taboo[c], inv_cov)
                    accept = d > sigma * taboo_d[c] * scale
                    if not accept:
                        break
                if accept:
                    x = np.vstack((x, elt_x[k]))
                    y = np.hstack((y, elt_y[k]))
                    z = np.vstack((z, elt_z[k]))
                    sigmas = np.hstack((sigmas, elt_s[k]))
        return x, y, z, sigmas

    def generate_taboo_point(self, superior_mean, superior_d, archive_x, archive_y, archive_d, elt_y):
        # Generate taboo points
        taboo = np.copy(superior_mean)  # Center of fitter subpopulations
        taboo_d = np.copy(superior_d)  # The normalized taboo distance of fitter subpopulations
        for i in range(len(archive_x)):  # Consider fitter archived points as taboo points
            base = np.ones((self.n_elt,))[archive_y[i] < elt_y]
            if len(base) == self.n_elt:
                taboo = np.vstack((taboo, archive_x[i]))
                taboo_d = np.hstack((taboo_d, archive_d[i]))
        return taboo, taboo_d

    def index_critical_points(self, x, sigma, max_eig_sqrt, taboo_points, taboo_points_d):
        # Equation (12)
        if len(taboo_points) > 0:
            mu1 = np.sqrt(np.sum(np.power(x - taboo_points, 2), 1)) / (max_eig_sqrt * sigma)  # L / (mu_1 * sigma_mean)
            criticality = norm.cdf(mu1 + taboo_points_d) - norm.cdf(mu1 - taboo_points_d)
            return np.argsort(-1 * criticality)[np.sort(-1 * criticality) < -1 * self.c_threshold]
        return np.empty((0,))

    def sample(self, mean, sigma, sqrt_cov, inv_cov, taboo, taboo_d, c_index, args=None):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        z = np.empty((self.n_individuals, self.ndim_problem))
        sigmas = np.empty((self.n_individuals,))
        scale = 1  # for temporary shrinkage of the taboo regions
        for n in range(self.n_individuals):
            accept = False
            while not accept:
                sigmas[n] = sigma * np.exp(self.rng_optimization.standard_normal() * self.tau)  # for Fig. 2. (R1)
                z[n] = np.dot(sqrt_cov, self.rng_optimization.standard_normal((self.ndim_problem,)))  # for Fig. 2. (R2)
                z[n] = sigmas[n] * z[n]  # for Fig. 2. (R3)
                x[n] = mean + z[n]  # for Fig. 2. (R4)
                if len(c_index) == 0:
                    break
                for c in c_index:
                    d = mahalanobis_distance(x[n], taboo[c], inv_cov)
                    accept = d > sigma * taboo_d[c] * scale
                    if not accept:  # reject
                        scale *= self.c_red  # Temporary shrink the size of the taboo regions
                        break
            y[n] = self._evaluate_fitness(x[n], args)
        return x, y, z, sigmas, scale

    def _update_distribution(self, x=None, z=None, y=None, cov=None, sigma=None, sigmas=None):
        # Update parameters of the subpopulation using Equation 6
        order = np.argsort(y)
        mean = np.sum(self._w.reshape(self.n_parents, 1) * x[order[:self.n_parents]], 0)
        c = np.zeros((self.ndim_problem, self.ndim_problem))
        for i in range(self.n_parents):
            m1, m2 = np.meshgrid(z[order[i]], z[order[i]])
            c += self._w[i] * (m2 * m1)
        cov = (1 - self.c_cov) * cov + self.c_cov * c
        cov = (cov + cov.T) / 2  # A symmetric matrix
        sigma *= np.exp(np.dot(self._w, np.log(sigmas[:self.n_parents]))) / np.exp(np.mean(np.log(sigmas)))
        return mean, cov, sigma

    def iterate(self, means=None, sigmas=None, cov=None, means_d=None,
                archive_x=None, archive_y=None, archive_d=None, args=None):
        n1 = 120 + int(30 * self.ndim_problem / self.n_individuals)  # no improvement
        n2 = 10 + int(30 * self.ndim_problem / self.n_individuals)  # stagnation
        cov_s = {k: cov for k in range(self.n_s)}
        elt_x = {k: np.tile(means[k], (self.n_elt, 1)) for k in range(self.n_s)}
        elt_y = {k: np.ones((self.n_elt,)) * self._best_so_far_y_bak for k in range(self.n_s)}
        elt_z = {k: np.zeros((self.n_elt, self.ndim_problem)) for k in range(self.n_s)}
        elt_s = {k: np.ones((self.n_elt,)) * np.mean(sigmas) for k in range(self.n_s)}
        best_y_ne = {k: np.empty((0,)) for k in range(self.n_s)}
        median_y_ne = {k: np.empty((0,)) for k in range(self.n_s)}
        superior_index = {k: np.arange(k) for k in range(self.n_s)}
        best_x = np.copy(means)
        best_y = np.ones((self.n_s,)) * self._best_so_far_y_bak
        ap = np.arange(self.n_s)  # activating subpopulations
        tp = []  # terminated subpopulations
        n_u = np.ones((self.n_s,))  # used iteration number

        n_iteration = 0
        while len(ap) > 0:
            for m in ap:
                # Termination subpopulation condition
                flag = [np.linalg.cond(cov_s[m]) >= self.n_condition, False, False, False]
                if n_iteration >= n2:
                    flag[1] = (np.max(best_y_ne[m][-n2:]) - np.min(best_y_ne[m][-n2:])) < self.fitness_threshold
                if n_iteration >= n1:
                    flag[2] = (np.median(best_y_ne[m][-20:]) - np.median(best_y_ne[m][-n1:(-n1 + 20)])) >= 0
                    flag[3] = (np.median(median_y_ne[m][-20:]) - np.median(median_y_ne[m][-n1:(-n1 + 20)])) >= 0
                if any(flag):
                    means[m] = best_x[m]
                    tp.append(m)
                    n_u[m] = n_iteration + 1
                    continue

                best_x[m], best_y[m], \
                    means[m], cov_s[m], sigmas[m], elt_x[m], elt_y[m], elt_z[m], elt_s[m], \
                    best_y_ne[m], median_y_ne[m] = \
                    self.iterate_subpopulation(
                        means[m], cov_s[m], sigmas[m], elt_x[m], elt_y[m], elt_z[m], elt_s[m],
                        best_y_ne[m], median_y_ne[m],
                        means[superior_index[m]], means_d[superior_index[m]],
                        archive_x, archive_y, archive_d, args)

                if self._check_terminations():
                    return best_x, best_y, np.mean(n_u)

            # Find non-terminated subpopulations
            ap = np.setdiff1d(ap, tp)
            if len(ap) > 0:
                ap = ap[np.argsort(best_y[ap])]
                base = np.arange(self.n_s)
                superior_index.update({m: base[best_y < best_y[m]] for m in ap})
            n_iteration += 1
            n1 = 120 + int(0.2 * n_iteration + 30 * self.ndim_problem / self.n_individuals)
        return best_x, best_y, np.mean(n_u)

    def optimize(self, fitness_function=None, args=None):
        fitness = ES.optimize(self, fitness_function)
        archive_x, archive_y, archive_d = self.initialize(args)
        while True:
            means, sigmas, cov, means_d = self.initialize_subpopulation(archive_x, archive_d)  # (Algorithm 2)
            best_x, best_y, n_u = self.iterate(means, sigmas, cov, means_d, archive_x, archive_y, archive_d, args)
            # Update archive (Algorithm 1)
            archive_x, archive_y, archive_d = self.update_archive(best_x, best_y, archive_x, archive_y, archive_d)
            if self.record_fitness:
                fitness.extend(best_y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(best_y)
            self.update_population_size(n_u)  # Restart (Equations 8)
        return self._collect_results(fitness)



import numpy as np
import matplotlib.pyplot as plt
from celluloid import Camera

from pypop7.benchmarks.utils import generate_xyz
from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.lmcmaes import LMCMAES


def plot_contour(func, x, y, levels=None):
    x, y, z = generate_xyz(func, x, y, 500)
    plt.contourf(x, y, z, levels, cmap='plasma')
    plt.xlabel('x')
    plt.ylabel('y')


def shi_cd(x):  # fitness function from https://arxiv.org/abs/1610.00040
    return 7 * np.power(x[0], 2) + 6 * x[0] * x[1] + 8 * np.power(x[1], 2)


class LMCMAES1(LMCMAES):
    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, x, p_c, s, vm, pm, b, d, y = self.initialize(args)
        x_data, mean_data = [], []
        while True:
            y_bak = np.copy(y)
            x, y = self.iterate(mean, x, pm, vm, y, b, args)  # sample and evaluate offspring population
            x_data.append(np.copy(x))
            mean_data.append(np.copy(mean))
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p_c, s, vm, pm, b, d = self._update_distribution(mean, x, p_c, s, vm, pm, b, d, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, x, p_c, s, vm, pm, b, d, y = self.restart_initialize(args, mean, x, p_c, s, vm, pm, b, d, y)
        results = self._collect_results(fitness, mean)
        results['x_data'] = x_data
        results['mean_data'] = mean_data
        return results


if __name__ == '__main__':
    ndim_problem = 2
    pro = {'fitness_function': shi_cd,
           'ndim_problem': ndim_problem}
    opt = {'fitness_threshold': 1e-10,
           'seed_rng': 0,
           'x': np.array([7., -8.]),  # mean
           'sigma': 0.1,
           'verbose_frequency': 5,
           'n_individuals': 250,
           'is_restart': False}
    solver = LMCMAES1(pro, opt)
    res = solver.optimize()
    fig = plt.figure()
    plt.rcParams['font.family'] = 'Times New Roman'
    plt.rcParams['font.size'] = '12'
    camera = Camera(fig)
    for i in range(len(res['mean_data'])):
        plt.xlim([-10, 10])
        plt.ylim([-10, 10])
        plt.xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plt.yticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plot_contour(shi_cd, [-10, 10], [-10, 10], [0, 10, 100, 500, 1000, 2000])
        plt.scatter(res['x_data'][i][:, 0], res['x_data'][i][:, 1], c='white', s=3)
        plt.scatter(res['mean_data'][i][0], res['mean_data'][i][1], c='magenta', s=12)
        camera.snap()
    animation = camera.animate()
    animation.save('demo_lmcmaes.gif')



"""PyPop7 is a Pure-PYthon library of POPulation-based OPtimization for single-objective, real-parameter, black-box
    problems (currently actively developed). Its main goal is to provide a unified interface and elegant implementations
    for Black-Box Optimization (BBO), particularly population-based optimizers, in order to facilitate research
    repeatability and also real-world applications.

    More specifically, for alleviating the notorious curse of dimensionality of BBO (almost based on iterative
    sampling), the primary focus of PyPop7 is to cover their State-Of-The-Art (SOTA) implementations for
    Large-Scale Optimization (LSO), though many of their other versions and variants are also included here
    (for benchmarking/mixing purpose, and sometimes even for practical purpose).
"""



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class DS(Optimizer):
    """Direct Search (DS).

    This is the **base** class for all DS classes.

    .. note:: Its three methods (`initialize`, `iterate`, `optimize`) should be implemented by
       its subclass.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with two particular settings (`keys`):
                * 'x'     - initial (starting) point (`array_like`),
                * 'sigma' - initial (global) step-size (`float`).

    Attributes
    ----------
    x     : `array_like`
            initial (starting) point.
    sigma : `float`
            (global) step-size.

    Methods
    -------

    References
    ----------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization. MIT Press.
    https://algorithmsbook.com/optimization/
    (See Chapter 7: Direct Methods for details.)

    Audet, C. and Hare, W., 2017. Derivative-free and blackbox optimization.
    Berlin: Springer International Publishing.
    https://link.springer.com/book/10.1007/978-3-319-68913-5

    Torczon, V., 1997.
    On the convergence of pattern search algorithms.
    SIAM Journal on Optimization, 7(1), pp.1-25.
    https://epubs.siam.org/doi/abs/10.1137/S1052623493250780

    `Wright, M.H. <https://www.informs.org/Explore/History-of-O.R.-Excellence/Biographical-Profiles/Wright-Margaret-H>`_
    , 1996. Direct search methods: Once scorned, now respectable.
    Pitman Research Notes in Mathematics Series, pp.191-208.
    https://nyuscholars.nyu.edu/en/publications/direct-search-methods-once-scorned-now-respectable

    Nelder, J.A. and Mead, R., 1965.
    A simplex method for function minimization.
    The Computer Journal, 7(4), pp.308-313.
    https://academic.oup.com/comjnl/article-abstract/7/4/308/354237

    Hooke, R. and Jeeves, T.A., 1961.
    “Direct search” solution of numerical and statistical problems.
    Journal of the ACM, 8(2), pp.212-229.
    https://dl.acm.org/doi/10.1145/321062.321069

    Fermi, E. and Metropolis N., 1952.
    Numerical solution of a minimum problem.
    Los Alamos Scientific Lab., Los Alamos, NM.
    https://www.osti.gov/servlets/purl/4377177
    """
    def __init__(self, problem, options):
        """Initialize the parameter settings of the DS class.

        Parameters
        ----------
        problem : dict
                  problem arguments.
        options : dict
                  optimizer options.
        """
        Optimizer.__init__(self, problem, options)
        self.x = options.get('x')  # initial (starting) point
        assert np.array(self.x).shape == (self.ndim_problem,),\
            f'`np.array(self.x).shape` == ({np.array(self.x).shape},), but should == ({self.ndim_problem},).'
        self.sigma = options.get('sigma')  # initial (global) step-size
        assert self.sigma > 0.0, f'`self.sigma` == {self.sigma}, but should > 0.0.'
        self._n_generations = 0  # number of generations
        # for restart
        self.n_restart = 0  # number of restarts
        self.sigma_threshold = options.get('sigma_threshold', 1e-10)  # stopping threshold of sigma for restart
        # maximal generation number of fitness stagnation for restart
        self.stagnation = options.get('stagnation', np.maximum(32, self.ndim_problem))
        self.fitness_diff = options.get('fitness_diff', 1e-10)  # stopping threshold of fitness difference for restart
        self._sigma_bak = np.copy(self.sigma)  # bak for restart
        self._fitness_list = [self.best_so_far_y]  # to store `best_so_far_y` generated in each generation

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _initialize_x(self, is_restart=False):
        if is_restart or (self.x is None):
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        return x

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness):
        results = Optimizer._collect_results(self, fitness)
        results['sigma'] = self.sigma
        results['n_restart'] = self.n_restart
        results['_n_generations'] = self._n_generations
        return results



import numpy as np
import matplotlib.pyplot as plt
from celluloid import Camera

from pypop7.optimizers.cem.cem import CEM
from pypop7.optimizers.cem.scem import SCEM

plt.rcParams['font.family'] = 'Times New Roman'
plt.rcParams['font.size'] = '12'


def _generate_xyz(func, x, y, num):
    x, y = np.array(x), np.array(y)
    if x.size == 2:
        x = np.linspace(x[0], x[1], num)
    if y.size == 2:
        y = np.linspace(y[0], y[1], num)
    x, y = np.meshgrid(x, y)
    z = np.empty(x.shape)
    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            z[i, j] = func([x[i, j], y[i, j]])
    return x, y, z


def plot_contour(func, x, y, levels=None):
    x, y, z = _generate_xyz(func, x, y, 500)
    plt.contourf(x, y, z, levels, cmap='bone')
    plt.xlabel('x')
    plt.ylabel('y')


def shi_cd(x):
    return 7 * np.power(x[0], 2) + 6 * x[0] * x[1] + 8 * np.power(x[1], 2)


class SCEM1(SCEM):
    def __init__(self, problem, options):
        SCEM.__init__(self, problem, options)
        self.x_data = []
        self.mean_data = []

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = CEM.optimize(self, fitness_function)
        mean, x, y = self.initialize()
        while True:
            x, y = self.iterate(mean, x, y)
            self.x_data.append(np.copy(x))
            self.mean_data.append(np.copy(mean))
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            mean = self._update_parameters(mean, x, y)
        results = self._collect_results(fitness, mean)
        results['x_data'] = self.x_data
        results['mean_data'] = self.mean_data
        return results


if __name__ == '__main__':
    ndim_problem = 2
    pro = {'fitness_function': shi_cd,
           'ndim_problem': ndim_problem}
    opt = {'fitness_threshold': 1e-10,
           'seed_rng': 0,
           'x': np.array([7., -8.]),  # mean
           'sigma': 1.0,
           'n_parents': 10,
           'n_individuals': 100,
           'verbose_frequency': 200,
           'max_function_evaluations': 100 * 100}
    solver = SCEM1(pro, opt)
    res = solver.optimize()
    fig = plt.figure()
    camera = Camera(fig)
    for i in range(10):
        plt.xlim([-10, 10])
        plt.ylim([-10, 10])
        plt.xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plt.yticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plot_contour(shi_cd, [-10, 10], [-10, 10], [0, 10, 100, 500, 1000, 2000])
        plt.scatter(res['x_data'][i][:, 0], res['x_data'][i][:, 1], c='limegreen', s=3)
        plt.scatter(res['mean_data'][i][0], res['mean_data'][i][1], c='magenta', s=12)
        plt.pause(0.05)
        camera.snap()
    animation = camera.animate()
    animation.save('demo.gif')



import numpy as np

from pypop7.optimizers.es.es import ES


class R1ES(ES):
    """Rank-One Evolution Strategy (R1ES).

    .. note:: `R1ES` is a **low-rank** version of `CMA-ES` specifically designed for large-scale black-box optimization
       (LSBBO) by Li and `Zhang <https://tinyurl.com/32hsbx28>`_. It often works well when there is a dominated search
       direction embedded in the subspace. For more complex landscapes (e.g., there are multiple promising search
       directions), other competitive LSBBO variants (e.g., `Rm-ES`, `LM-CMA`, `LM-MA-ES`) may be more preferred.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - generation frequency of printing verbose info (`int`, default: `10`);
              and with four particular settings (`keys`):
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`),
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`).

                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`, default:
                  `4 + int(3*np.log(self.ndim_problem))`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`, default:
                  `int(self.n_individuals / 2)`),
                * 'c_cov'         - learning rate of low-rank covariance matrix (`float`, default:
                  `1.0/(3.0*np.sqrt(self.ndim_problem) + 5.0)`),
                * 'c'             - learning rate of evolution path (`float`, default: `2.0/(self.ndim_problem + 7.0)`),
                * 'c_s'           - learning rate of cumulative (global) step-size adaptation (`float`, default: `0.3`),
                * 'q_star'        - baseline of cumulative (global) step-size adaptation (`float`, default: `0.3`)，
                * 'd_sigma'       - change factor of cumulative (global) step-size adaptation (`float`, default: `1.0`).

    Examples
    --------
    Use the ES optimizer `R1ES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.r1es import R1ES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> r1es = R1ES(problem, options)  # initialize the optimizer class
       >>> results = r1es.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"R1ES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       R1ES: 5000, 0.0005057573524421161

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/2aywpp2p>`_ for more details.

    Attributes
    ----------
    n_individuals   : `int`
                      number of offspring (λ: lambda), offspring population size.
    n_parents       : `int`
                      number of parents (μ: mu), parental population size.
    mean            : `array_like`
                      initial (starting) point, mean of Gaussian search distribution.
    sigma           : `float`
                      initial global step-size (σ), mutation strength (`float`).
    c_cov           : `float`
                      learning rate of low-rank covariance matrix.
    c               : `float`
                      learning rate of evolution path.
    c_s             : `float`
                      learning rate of cumulative (global) step-size adaptation.
    q_star          : `float`
                      baseline of cumulative (global) step-size adaptation.
    d_sigma         : `float`
                      change factor of cumulative (global) step-size adaptation.

    References
    ----------
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_cov = options.get('c_cov', 1.0/(3.0*np.sqrt(self.ndim_problem) + 5.0))  # for Line 5 in Algorithm 1
        self.c = options.get('c', 2.0/(self.ndim_problem + 7.0))  # for Line 12 in Algorithm 1 (c_c)
        self.c_s = options.get('c_s', 0.3)  # for Line 15 in Algorithm 1
        self.q_star = options.get('q_star', 0.3)  # for Line 15 in Algorithm 1
        self.d_sigma = options.get('d_sigma', 1.0)  # for Line 16 in Algorithm 1
        self._x_1 = np.sqrt(1.0 - self.c_cov)  # for Line 5 in Algorithm 1
        self._x_2 = np.sqrt(self.c_cov)  # for Line 5 in Algorithm 1
        self._p_1 = 1.0 - self.c  # for Line 12 in Algorithm 1
        self._p_2 = None  # for Line 12 in Algorithm 1
        self._rr = None  # for rank-based success rule (RSR)

    def initialize(self, args=None, is_restart=False):
        self._p_2 = np.sqrt(self.c*(2.0 - self.c)*self._mu_eff)
        self._rr = np.arange(self.n_parents*2) + 1
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p = np.zeros((self.ndim_problem,))  # principal search direction
        s = 0.0  # cumulative rank rate
        y = np.tile(self._evaluate_fitness(mean, args), (self.n_individuals,))  # fitness
        return x, mean, p, s, y

    def iterate(self, x=None, mean=None, p=None, y=None, args=None):
        for k in range(self.n_individuals):  # for Line 3 in Algorithm 1
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))  # for Line 4 in Algorithm 1
            r = self.rng_optimization.standard_normal()  # for Line 4 in Algorithm 1
            # for Line 5 in Algorithm 1
            x[k] = mean + self.sigma*(self._x_1*z + self._x_2*r*p)
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, p=None, s=None, y=None, y_bak=None):
        order = np.argsort(y)
        y.sort()  # for Line 10 in Algorithm 1
        # for Line 11 in Algorithm 1
        mean_w = np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):
            mean_w += self._w[k]*x[order[k]]
        p = self._p_1*p + self._p_2*(mean_w - mean) / self.sigma  # for Line 12 in Algorithm 1
        mean = mean_w  # for Line 11 in Algorithm 1
        # for rank-based adaptation of mutation strength
        r = np.argsort(np.hstack((y_bak[:self.n_parents], y[:self.n_parents])))
        rr = self._rr[r < self.n_parents] - self._rr[r >= self.n_parents]
        q = np.dot(self._w, rr) / self.n_parents  # for Line 14 in Algorithm 1
        s = (1 - self.c_s)*s + self.c_s*(q - self.q_star)  # for Line 15 in Algorithm 1
        self.sigma *= np.exp(s / self.d_sigma)  # for Line 16 in Algorithm 1
        return mean, p, s

    def restart_initialize(self, args=None, x=None, mean=None, p=None, s=None, y=None, fitness=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            x, mean, p, s, y = self.initialize(args, is_restart)
            fitness.append(y[0])
            self.d_sigma *= 2.0
        return x, mean, p, s, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, s, y = self.initialize(args)
        fitness.append(y[0])
        while True:
            y_bak = np.copy(y)  # for Line 13 in Algorithm 1
            x, y = self.iterate(x, mean, p, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, s = self._update_distribution(x, mean, p, s, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, p, s, y = self.restart_initialize(args, x, mean, p, s, y, fitness)
        results = self._collect_results(fitness, mean)
        results['p'] = p
        results['s'] = s
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class PSO(Optimizer):
    """Particle Swarm Optimizer (PSO).

    This is the **base** (abstract) class for all `PSO` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand.

    .. note:: `PSO` is a very popular family of *swarm*-based search algorithms, proposed together by an electrical
       engineer (Russell C. Eberhart) and a psychologist (James Kennedy), recipients of `Evolutionary Computation
       Pioneer Award 2012 <https://tinyurl.com/456as566>`_. Its underlying motivation comes from very interesting
       collective behaviors (e.g. `flocking <https://dl.acm.org/doi/10.1145/37402.37406>`_) observed from social
       animals, which are regarded as one particular form of *intelligence* or *emergence* by many scientists.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose information during optimization (`bool`, default:
                  `True`),
                * 'verbose_frequency'        - generation frequency of printing verbose information (`int`, default:
                  `10`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, number of particles (`int`, default: `20`),
                * 'cognition'     - cognitive learning rate (`float`, default: `2.0`),
                * 'society'       - social learning rate (`float`, default: `2.0`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Attributes
    ----------
    n_individuals : `int`
                    swarm (population) size, number of particles.
    cognition     : `float`
                    cognitive learning rate, acceleration coefficient.
    society       : `float`
                    social learning rate, acceleration coefficient.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.

    Methods
    -------

    References
    ----------
    Bonyadi, M.R. and Michalewicz, Z., 2017.
    Particle swarm optimization for single objective continuous space problems: A review.
    Evolutionary Computation, 25(1), pp.1-54.
    https://direct.mit.edu/evco/article-abstract/25/1/1/1040/Particle-Swarm-Optimization-for-Single-Objective

    Poli, R., Kennedy, J. and Blackwell, T., 2007.
    Particle swarm optimization.
    Swarm Intelligence, 1(1), pp.33-57.
    https://link.springer.com/article/10.1007/s11721-007-0002-0

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146

    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # swarm (population) size, number of particles
            self.n_individuals = 20
        self.cognition = options.get('cognition', 2.0)  # cognitive learning rate
        self.society = options.get('society', 2.0)  # social learning rate
        self.max_ratio_v = options.get('max_ratio_v', 0.2)  # maximal ratio of velocities
        self._max_v = self.max_ratio_v*(self.upper_boundary - self.lower_boundary)  # maximal velocity
        self._min_v = -self._max_v  # minimal velocity
        self._topology = None  # neighbors topology of social learning
        self._n_generations = 0  # number of generations
        # linearly decreasing inertia weights introduced in [Shi&Eberhart, 1998, WCCI]
        self._max_generations = np.ceil(self.max_function_evaluations/self.n_individuals)
        self._w = 0.9 - 0.5*(np.arange(self._max_generations) + 1)/self._max_generations  # from 0.9 to 0.4
        self._swarm_shape = (self.n_individuals, self.ndim_problem)

    def initialize(self, args=None):
        v = self.rng_initialization.uniform(self._min_v, self._max_v, size=self._swarm_shape)  # velocities
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=self._swarm_shape)  # positions
        y = np.empty((self.n_individuals,))  # fitness
        p_x, p_y = np.copy(x), np.copy(y)  # personally previous-best positions and fitness
        n_x = np.copy(x)  # neighborly previous-best positions
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            y[i] = self._evaluate_fitness(x[i], args)
        p_y = np.copy(y)
        return v, x, y, p_x, p_y, n_x

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        self._n_generations += 1  # to be implemented by its all subclasses
        raise NotImplementedError

    def _print_verbose_info(self, y=None):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        v, x, y, p_x, p_y, n_x = self.initialize(args)
        if self.record_fitness:
            fitness.extend(y)
        while True:
            v, x, y, p_x, p_y, n_x = self.iterate(v, x, y, p_x, p_y, n_x, args)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._print_verbose_info(y)
        return self._collect_results(fitness)



"""Repeat Figure 1 (left sub-figure) from the following paper:
    Akimoto, Y., Auger, A. and Hansen, N., 2014, July.
    Comparison-based natural gradient optimization in high dimension.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 373-380). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598258

    See the official Python version from Akimoto:
    https://gist.github.com/youheiakimoto/08b95b52dfbf8832afc71dfff3aed6c8
"""
import pickle
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.vdcma import VDCMA
from pypop7.benchmarks.base_functions import sphere

sns.set_theme(style='darkgrid')


def _rot_cigar(x):
    u = np.ones((len(x),)) / np.sqrt(len(x))  # unit vector
    y = np.power(10, 6) * sphere(x) + (1 - np.power(10, 6)) * np.power(np.dot(x, u), 2)
    return y


def _ell_cig(x):
    dim = len(x)
    d_ell = np.zeros((dim, dim))
    for i in range(dim):
        d_ell[i][i] = np.power(10, 3 * i / dim)  # diagonal matrix
    y = _rot_cigar(np.dot(d_ell, x))
    return y


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, problem_dim):
    plt.figure()
    result = read_pickle(function, problem_dim)
    plt.yscale('log')
    plt.xlim([0, 3.5e4])
    plt.xticks([1e4, 2e4, 3e4])
    plt.ylim([1e-10, 1e13])
    plt.yticks([1e-10, 1e-5, 1e0, 1e5, 1e10])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='b', label=r'$min_i$f($x_i$)')
    plt.plot(result['fitness'][:, 0], result['sigma'], color='g', label=r'$\sigma$')
    plt.xlabel("No. Func. Evals")
    plt.legend(loc="upper right", bbox_to_anchor=(1, 1))
    plt.show()


class Fig1(VDCMA):
    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness, sigma = ES.optimize(self, fitness_function), []
        d, v, p_s, p_c, z, zz, x, mean, y = self.initialize()
        while True:
            z, zz, x, y = self.iterate(d, z, zz, x, mean, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
                sigma.extend(np.ones((len(y),)) * self.sigma)
            if self._check_terminations():
                break
            mean, p_s, p_c, v, d = self._update_distribution(d, v, p_s, p_c, zz, x, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                d, v, p_s, p_c, z, zz, x, mean, y = self.restart_initialize(d, v, p_s, p_c, z, zz, x, mean, y)
        res = self._collect_results(fitness, mean)
        res['d'] = d
        res['v'] = v
        res['sigma'] = sigma
        return res


if __name__ == '__main__':
    # Fig. 1 (left sub-figure)
    ndim_problem = 50
    print('*' * 7 + ' ' + _ell_cig.__name__ + ' ' + '*' * 7)
    problem = {'fitness_function': _ell_cig,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5 * np.ones((ndim_problem,)),
               'upper_boundary': 5 * np.ones((ndim_problem,))}
    options = {'fitness_threshold': 1e-10,
               'seed_rng': 0,  # not given in the original paper
               'mean': 3 + 2 * np.random.default_rng(0).standard_normal((ndim_problem,)),
               'sigma': 2,
               'verbose_frequency': 200,
               'record_fitness': True,
               'record_fitness_frequency': 1}
    solver = Fig1(problem, options)
    results = solver.optimize()
    write_pickle(_ell_cig.__name__, ndim_problem, results)
    plot(_ell_cig.__name__, ndim_problem)



"""Repeat Fig.2 in this paper
    (1+1)-Active-CMA-ES (OPOA).
    Reference
    ---------
    Arnold, D.V. and Hansen, N., 2010, July.
    Active covariance matrix adaptation for the (1+1)-CMA-ES.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 385-392). ACM.
    https://dl.acm.org/doi/abs/10.1145/1830483.1830556
    """
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle

from pypop7.optimizers.es.opoa import OPOA as Solver
from pypop7.benchmarks.base_functions import sphere, ellipsoid, cigar, discus, cigar_discus

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(functions, evaluations, problem_dim):
    plt.xticks(rotation=30)
    plt.bar(functions, evaluations, color='g', width=0.5)
    if problem_dim == 2:
        plt.ylim([0, 6e2])
        plt.yticks([0, 2e2, 4e2, 6e2])
    elif problem_dim == 4:
        plt.ylim([0, 1500])
        plt.yticks([0, 5e2, 1e3, 1.5e3])
    elif problem_dim == 10:
        plt.ylim([0, 7.2e3])
        plt.yticks([0, 2.4e3, 4.8e3, 7.2e3])
    elif problem_dim == 20:
        plt.ylim([0, 2.4e4])
        plt.yticks([0, 8e3, 1.6e4, 2.4e4])
    elif problem_dim == 40:
        plt.ylim([0, 8.4e4])
        plt.yticks([0, 2.8e4, 5.6e4, 8.4e4])
    plt.title(" Dimension: " + str(problem_dim))
    plt.show()


if __name__ == '__main__':
    functions = [sphere, ellipsoid, cigar, discus, cigar_discus]
    function_name = ['sphere', 'ellipsoid', 'cigar', 'discus', 'cigdis']
    for d in [10, 20, 40]:
        evaluations = []
        for f in functions:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'max_function_evaluations': 2e6,
                       'fitness_threshold': 1e-9,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 0.1,
                       'verbose_frequency': 2000,
                       'record_fitness': True,
                       'is_restart': False,
                       'record_fitness_frequency': 1}
            solver = Solver(problem, options)
            results = solver.optimize()
            evaluations.append(results['n_function_evaluations'])
        plot(function_name, evaluations, d)



"""Repeat the following paper for `AEMNA`:
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    (See Chapter 8 Experimental Results in Function Optimization with EDAs in Continuous Domain.)

    Surprisingly, our code reported *much better* results than the original paper. However, given that it shows
    very similar performance with its same kind `EMNA`, we still argue that the repeatability of `AEMNA` can be
    well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.eda.aemna import AEMNA


if __name__ == '__main__':
    ndim_problem = 10

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0,  # undefined in the original paper
               'verbose_frequency': 1000}
    aemna = AEMNA(problem, options)
    results = aemna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 7.517279751033313
    # vs 3263.0010 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0,  # undefined in the original paper
               'verbose_frequency': 1000}
    aemna = AEMNA(problem, options)
    results = aemna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.7944761831988641
    # vs 12.9407 (from the original paper)



"""Repeat Fig.6, Fig.7 in paper
    Cholesky-CMA-ES (CCMAES, (μ/μ_w,λ)-Cholesky-CMA-ES).
    Reference
    ---------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 4 for details.)
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle

from pypop7.optimizers.es.ccmaes import CCMAES as Solver
from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, problem_dim):
    plt.figure()
    result = read_pickle(function, problem_dim)
    plt.yscale('log')
    if function == 'ellipsoid':
        plt.ylim([1e-35, 1e5])
        plt.yticks([1e-35, 1e-30, 1e-25, 1e-20, 1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 3e4])
        plt.xticks([0, 5e3, 1e4, 1.5e4, 2e4, 2.5e4, 3e4])
    elif function == 'rosenbrock':
        plt.ylim([1e-20, 1e5])
        plt.yticks([1e-20, 1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.xlim([0, 3e4])
        plt.xticks([0, 5e3, 1e4, 1.5e4, 2e4, 2.5e4, 3e4])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='black')
    plt.xlabel("objective function evaluations")
    plt.ylabel("objective function value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    for f in [ellipsoid, rosenbrock]:
        for d in [20]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'fitness_threshold': 1e-10,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'x': 0.2 * np.ones((d,)),  # mean
                       'sigma': 0.2/3,
                       'verbose_frequency': 2000,
                       'record_fitness': True,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            if f.__name__ == 'ellipsoid':
                options['fitness_threshold'] = 1e-30
                options['max_function_evaluations'] = 3e4
            elif f.__name__ == 'rosenbrock':
                options['fitness_threshold'] = 1e-15
                options['max_function_evaluations'] = 3e4
            solver = Solver(problem, options)
            results = solver.optimize()
            write_pickle(f.__name__, d, results)
            plot(f.__name__, d)



from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.res import RES
from pypop7.optimizers.es.ssaes import SSAES
from pypop7.optimizers.es.dsaes import DSAES
from pypop7.optimizers.es.csaes import CSAES
from pypop7.optimizers.es.saes import SAES
from pypop7.optimizers.es.sepcmaes import SEPCMAES
from pypop7.optimizers.es.maes import MAES
from pypop7.optimizers.es.fmaes import FMAES
from pypop7.optimizers.es.r1es import R1ES
from pypop7.optimizers.es.lmcmaes import LMCMAES
from pypop7.optimizers.es.lmcma import LMCMA
from pypop7.optimizers.es.lmmaes import LMMAES


__all__ = [ES,  # base (abstract) class
           RES, SSAES, DSAES, CSAES, SAES,  # representative ES versions during development
           MAES, FMAES,  # modern ES versions often with state-of-the-art performance
           SEPCMAES, R1ES, LMCMAES, LMCMA, LMMAES]  # especially for large-scale black-box optimization



import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.eda.rpeda import RPEDA


if __name__ == '__main__':
    ndim_problem = 1000
    for f in [rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -100 * np.ones((ndim_problem,)),
                   'upper_boundary': 100 * np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 6e5,
                   'm': 1000,
                   'seed_rng': 0,
                   'verbose_frequency': 100,
                   'record_fitness': True}
        rpeda = RPEDA(problem, options)
        results = rpeda.optimize()
        print(f.__name__, results['best_so_far_y'])
        # rosenbrock 987.6557605512518 vs mean 1614.7 std 249.44 (from the original paper)



import numpy as np

from pypop7.optimizers.ga.ga import GA


class GENITOR(GA):
    """GENITOR.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `100`),

    Attributes
    ----------
    n_individuals : `int`
                    population size.

    References
    ----------
    Whitley, D., Dominic, S., Das, R. and Anderson, C.W., 1993.
    Genetic reinforcement learning for neurocontrol problems.
    Machine Learning, 13, pp.259-284.
    https://link.springer.com/article/10.1023/A:1022602019183
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.crossover_prob = options.get('crossover_prob', 0.5)  # crossover probability
        self._rank_prob = np.arange(self.n_individuals, 0, -1) - 1.0

    def initialize(self):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        crossover_probs = self.crossover_prob*np.ones((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i])
        self._n_generations += 1
        return x, y, crossover_probs

    def iterate(self, x=None, y=None, crossover_probs=None):
        order, xx, yy = np.argsort(y), None, None
        # use rank-based selection for two parents
        rank_prob = self._rank_prob/np.sum(self._rank_prob)
        offspring = self.rng_optimization.choice(order, size=2, replace=False, p=rank_prob)
        if self.rng_optimization.random() < crossover_probs[offspring[0]]:  # crossover
            # use intermediate crossover (not one-point crossover proposed in the original paper)
            xx = (x[offspring[0]] + x[offspring[1]])/2.0
            yy = self._evaluate_fitness(xx)
            if yy < y[order[-1]]:  # to replace the worst individual
                x[order[-1]], y[order[-1]] = xx, yy
                crossover_probs[offspring[0]] += 0.1
            else:
                crossover_probs[offspring[0]] -= 0.1
            crossover_probs[offspring[0]] = np.maximum(0.05, np.minimum(0.95, crossover_probs[offspring[0]]))
        else:  # mutation
            xx = np.copy(x[offspring[0]])  # offspring
            xx += self.rng_optimization.uniform(self.lower_boundary, self.upper_boundary)/10.0
            yy = self._evaluate_fitness(xx)
            if yy < y[order[-1]]:  # to replace the worst individual
                x[order[-1]], y[order[-1]] = xx, yy
        self._n_generations += 1
        return x, y, crossover_probs

    def optimize(self, fitness_function=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, crossover_probs = self.initialize()
        fitness.extend(y)
        while True:
            x, y, crossover_probs = self.iterate(x, y, crossover_probs)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._print_verbose_info(y)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.es.es import ES


class SEPCMAES(ES):
    """Separable Covariance Matrix Adaptation Evolution Strategy (SEPCMAES).

    .. note:: Only the **diagonal** elements of the full covariance matrix are saved explicitly,
       since all off-diagonal elements are not used for sep-CMA-ES.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with four particular settings (`keys`):
                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`),
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`).

    Examples
    --------
    Use the ES optimizer `SEPCMAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.sepcmaes import SEPCMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> sepcmaes = SEPCMAES(problem, options)  # initialize the optimizer class
       >>> results = sepcmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SEPCMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SEPCMAES: 5000, 0.010606023368122535

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mpjzv8yh>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring/descendants (λ: lambda), offspring population size.
    n_parents     : `int`
                    number of parents (μ: mu), parental population size.
    mean          : `array_like`
                    initial (starting) point, mean of Gaussian search distribution.
    sigma         : `float`
                    initial global step-size (σ), mutation strength (rate).

    References
    ----------
    Ros, R. and Hansen, N., 2008, September.
    A simple modification in CMA-ES achieving linear time and space complexity.
    In International Conference on Parallel Problem Solving from Nature (pp. 296-305).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-540-87700-4_30
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_c = options.get('c_c', 4.0/(self.ndim_problem + 4.0))
        self.options = options
        self.c_s = None
        self.c_cov = None
        self.d_sigma = None
        self._s_1 = None
        self._s_2 = None

    def _set_c_cov(self):
        c_cov = (1.0/self._mu_eff)*(2.0/np.power(self.ndim_problem + np.sqrt(2.0), 2)) + (
            (1.0 - 1.0/self._mu_eff)*np.minimum(1.0, (2.0*self._mu_eff - 1.0) / (
                np.power(self.ndim_problem + 2.0, 2) + self._mu_eff)))
        c_cov *= (self.ndim_problem + 2.0)/3.0  # for faster adaptation
        return c_cov

    def _set_d_sigma(self):
        d_sigma = np.maximum((self._mu_eff - 1.0)/(self.ndim_problem + 1.0) - 1.0, 0.0)
        return 1.0 + self.c_s + 2.0*np.sqrt(d_sigma)

    def initialize(self, is_restart=False):
        self.c_s = self.options.get('c_s', (self._mu_eff + 2.0)/(self.ndim_problem + self._mu_eff + 3.0))
        self.c_cov = self.options.get('c_cov', self._set_c_cov())
        self.d_sigma = self.options.get('d_sigma', self._set_d_sigma())
        self._s_1 = 1.0 - self.c_s
        self._s_2 = np.sqrt(self._mu_eff*self.c_s*(2.0 - self.c_s))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path for CSA
        p = np.zeros((self.ndim_problem,))  # evolution path for CMA
        c = np.ones((self.ndim_problem,))  # diagonal elements for covariance matrix
        d = np.ones((self.ndim_problem,))  # diagonal elements for covariance matrix
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return z, x, mean, s, p, c, d, y

    def iterate(self, z=None, x=None, mean=None, d=None, y=None, args=None):
        for k in range(self.n_individuals):  # Line 2 of Algorithm 1
            if self._check_terminations():
                return z, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            x[k] = mean + self.sigma*d*z[k]  # Line 3 of Algorithm 1
            y[k] = self._evaluate_fitness(x[k], args)
        return z, x, y

    def _update_distribution(self, z=None, x=None, s=None, p=None, c=None, d=None, y=None):
        order = np.argsort(y)
        zeros = np.zeros((self.ndim_problem,))
        z_w, mean, dz_w = np.copy(zeros), np.copy(zeros), np.copy(zeros)
        for k in range(self.n_parents):  # Line 4, 5 of Algorithm 1
            z_w += self._w[k]*z[order[k]]
            mean += self._w[k]*x[order[k]]  # update distribution mean
            dz = d*z[order[k]]
            dz_w += self._w[k]*dz*dz
        s = self._s_1*s + self._s_2*z_w  # Line 6 of Algorithm 1
        # Line 7 of Algorithm 1
        if (np.linalg.norm(s) / np.sqrt(1.0 - np.power(1.0 - self.c_s, 2.0*self._n_generations))) < (
                (1.4 + 2.0/(self.ndim_problem + 1.0))*self._e_chi):
            h = np.sqrt(self.c_c*(2.0 - self.c_c))*np.sqrt(self._mu_eff)*d*z_w
        else:
            h = 0
        p = (1.0 - self.c_c)*p + h  # Line 8 of Algorithm 1
        c = (1.0 - self.c_cov)*c + (1.0/self._mu_eff)*self.c_cov*p*p + (
                self.c_cov*(1.0 - 1.0/self._mu_eff)*dz_w)  # Line 9 of Algorithm 1
        # Line 10 of Algorithm 1
        self.sigma *= np.exp(self.c_s/self.d_sigma*(np.linalg.norm(s)/self._e_chi - 1.0))
        d = np.sqrt(c)  # Line 11 of Algorithm 1
        return mean, s, p, c, d

    def restart_initialize(self, z=None, x=None, mean=None, s=None, p=None, c=None, d=None, y=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            z, x, mean, s, p, c, d, y = self.initialize(is_restart)
            self._n_generations = 0
        return z, x, mean, s, p, c, d, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, x, mean, s, p, c, d, y = self.initialize()
        while True:
            self._n_generations += 1  # Line 1 of Algorithm 1
            z, x, y = self.iterate(z, x, mean, d, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, s, p, c, d = self._update_distribution(z, x, s, p, c, d, y)
            self._print_verbose_info(y)
            if self.is_restart:
                z, x, mean, s, p, c, d, y = self.restart_initialize(z, x, mean, s, p, c, d, y)
        results = self._collect_results(fitness, mean)
        results['s'] = s
        results['p'] = p
        results['d'] = d
        return results



"""Repeat experiments on four functions [Cigar, Ellipsoid, Rosenbrock, Sphere] from the following paper:
    Poland, J. and Zell, A., 2001, July.
    Main vector adaptation: A CMA variant with linear time and space complexity.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 1050-1055).
    https://dl.acm.org/doi/abs/10.5555/2955239.2955428
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from pypop7.benchmarks.base_functions import cigar, ellipsoid, rosenbrock, sphere
from pypop7.optimizers.es.mvaes import MVAES as Solver

sns.set_theme(style='darkgrid')


if __name__ == '__main__':
    ndim_problem = 20
    for f in [cigar, ellipsoid, rosenbrock, sphere]:
        plt.figure()
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem}
        options = {'max_function_evaluations': 5e5,
                   'fitness_threshold': 1e-10,
                   'seed_rng': 0,  # undefined in the original paper
                   'x': np.ones((ndim_problem,)),  # undefined in the original paper
                   'n_individuals': 10,
                   'n_parents': 1,
                   'sigma': 1,  # undefined in the original paper
                   'verbose_frequency': 200,
                   'record_fitness': True,
                   'record_fitness_frequency': 1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)
        plt.plot(results['fitness'][:, 0], results['fitness'][:, 1], 'k',
                 label=r'$\mu$=1, $\lambda$=10, no recombination')
        options['n_individuals'] = 35
        options['n_parents'] = 5
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)
        plt.plot(results['fitness'][:, 0], results['fitness'][:, 1], 'k--',
                 label=r'$\mu$=5, $\lambda$=35, recombination')
        plt.yscale('log')
        if f == cigar:
            plt.xticks(ticks=[0, 1e4, 2e4, 3e4, 4e4, 5e4])
            plt.xlim([0, 5e4])
        elif f == ellipsoid:
            plt.xticks(ticks=[0, 5e4, 1e5, 1.5e5, 2e5, 2.5e5, 3e5, 3.5e5])
            plt.xlim([0, 3.5e5])
        elif f == sphere:
            plt.xticks(ticks=[0, 1000, 2000, 3000, 4000, 5000, 6000])
            plt.xlim(([0, 6e3]))
        elif f == rosenbrock:
            plt.xticks(ticks=[0, 2e4, 4e4, 6e4, 8e4, 1e5])
            plt.xlim([0, 1e5])
        plt.ylim([1e-10, 1e4])
        plt.yticks(ticks=[1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4])
        plt.xlabel('function evaluations')
        plt.ylabel('fitness')
        plt.legend(loc='upper right', bbox_to_anchor=(1, 1))
        plt.title(f.__name__.capitalize())
        plt.show()
